__all__ = (
    "BaseMoMetric",
    "NObsMetric",
    "NObsNoSinglesMetric",
    "NNightsMetric",
    "ObsArcMetric",
    "DiscoveryMetric",
    "DiscoveryNChancesMetric",
    "DiscoveryNObsMetric",
    "DiscoveryTimeMetric",
    "DiscoveryDistanceMetric",
    "DiscoveryRadecMetric",
    "DiscoveryEclonlatMetric",
    "DiscoveryVelocityMetric",
    "ActivityOverTimeMetric",
    "ActivityOverPeriodMetric",
    "MagicDiscoveryMetric",
    "HighVelocityMetric",
    "HighVelocityNightsMetric",
    "LightcurveInversionAsteroidMetric",
    "ColorAsteroidMetric",
    "InstantaneousColorMetric",
    "LightcurveColorOuterMetric",
    "PeakVMagMetric",
    "KnownObjectsMetric",
)

import numpy as np

from .base_metric import BaseMetric


def _set_vis(sso_obs, snr_limit, snr_col, vis_col):
    if snr_limit is not None:
        vis = np.where(sso_obs[snr_col] >= snr_limit)[0]
    else:
        vis = np.where(sso_obs[vis_col] > 0)[0]
    return vis


class BaseMoMetric(BaseMetric):
    """Base class for the moving object metrics.
    Intended to be used with the Moving Object Slicer."""

    def __init__(
        self,
        cols=None,
        metric_name=None,
        units="#",
        badval=0,
        comment=None,
        child_metrics=None,
        app_mag_col="appMag",
        app_mag_v_col="appMagV",
        m5_col="fiveSigmaDepth",
        night_col="night",
        mjd_col="observationStartMJD",
        snr_col="SNR",
        vis_col="vis",
        ra_col="ra",
        dec_col="dec",
        seeing_col="seeingFwhmGeom",
        exp_time_col="visitExposureTime",
        filter_col="filter",
    ):
        # Set metric name.
        self.name = metric_name
        if self.name is None:
            self.name = self.__class__.__name__.replace("Metric", "", 1)
        # Set badval and units, leave space for 'comment' (tied to display_dict).
        self.badval = badval
        self.units = units
        self.comment = comment
        # Set some commonly used column names.
        self.m5_col = m5_col
        self.app_mag_col = app_mag_col
        self.app_mag_v_col = app_mag_v_col
        self.night_col = night_col
        self.mjd_col = mjd_col
        self.snr_col = snr_col
        self.vis_col = vis_col
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.seeing_col = seeing_col
        self.exp_time_col = exp_time_col
        self.filter_col = filter_col
        self.cols_req = [
            self.app_mag_col,
            self.m5_col,
            self.night_col,
            self.mjd_col,
            self.snr_col,
            self.vis_col,
        ]
        if cols is not None:
            for col in cols:
                self.cols_req.append(col)

        if child_metrics is None:
            try:
                if not isinstance(self.child_metrics, dict):
                    raise ValueError("self.child_metrics must be a dictionary (possibly empty)")
            except AttributeError:
                self.child_metrics = {}
                self.metric_dtype = "float"
        else:
            if not isinstance(child_metrics, dict):
                raise ValueError("childmetrics must be provided as a dictionary.")
            self.child_metrics = child_metrics
            self.metric_dtype = "object"
        self.shape = 1

    def run(self, sso_obs, orb, hval):
        """Calculate the metric value.

        Parameters
        ----------
        sso_obs: np.ndarray
            The input data to the metric (same as the parent metric).
        orb: np.ndarray
            The information about the orbit for which the metric is being calculated.
        hval : float
            The H value for which the metric is being calculated.

        Returns
        -------
        float or np.ndarray or dict
        """
        raise NotImplementedError


class BaseChildMetric(BaseMoMetric):
    """Base class for child metrics.

    Parameters
    ----------
    parentDiscoveryMetric: BaseMoMetric
        The 'parent' metric which generated the metric data used to calculate this 'child' metric.
    badval: float, optional
        Value to return when metric cannot be calculated.
    """

    def __init__(self, parent_discovery_metric, badval=0, **kwargs):
        super().__init__(badval=badval, **kwargs)
        self.parent_metric = parent_discovery_metric
        self.child_metrics = {}
        if "metricDtype" in kwargs:
            self.metric_dtype = kwargs["metricDtype"]
        else:
            self.metric_dtype = "float"

    def run(self, sso_obs, orb, hval, metric_values):
        """Calculate the child metric value.

        Parameters
        ----------
        sso_obs: np.ndarray
            The input data to the metric (same as the parent metric).
        orb: np.ndarray
            The information about the orbit for which the metric is being calculated.
        hval : float
            The H value for which the metric is being calculated.
        metric_values : dict or np.ndarray
            The return value from the parent metric.

        Returns
        -------
        float
        """
        raise NotImplementedError


class NObsMetric(BaseMoMetric):
    """
    Count the total number of observations where an SSobject was 'visible'.
    """

    def __init__(self, snr_limit=None, **kwargs):
        """
        @ snr_limit .. if snr_limit is None, this uses the _calcVis method/completeness
                      if snr_limit is not None, this uses that value as a cutoff instead.
        """
        super().__init__(**kwargs)
        self.snr_limit = snr_limit

    def run(self, sso_obs, orb, hval):
        if self.snr_limit is not None:
            vis = np.where(sso_obs[self.snr_col] >= self.snr_limit)[0]
            return vis.size
        else:
            vis = np.where(sso_obs[self.vis_col] > 0)[0]
            return vis.size


class NObsNoSinglesMetric(BaseMoMetric):
    """
    Count the number of observations for an SSobject, without singles.
    Don't include any observations where it was a single observation on a night.
    """

    def __init__(self, snr_limit=None, **kwargs):
        super().__init__(**kwargs)
        self.snr_limit = snr_limit

    def run(self, sso_obs, orb, hval):
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return 0
        nights = sso_obs[self.night_col][vis]
        nights = nights.astype("int")
        ncounts = np.bincount(nights)
        nobs = ncounts[np.where(ncounts > 1)].sum()
        return nobs


class NNightsMetric(BaseMoMetric):
    """Count the number of distinct nights an SSobject is observed."""

    def __init__(self, snr_limit=None, **kwargs):
        """
        @ snr_limit : if SNRlimit is None, this uses _calcVis method/completeness
                     else if snr_limit is not None, it uses that value as a cutoff.
        """
        super().__init__(**kwargs)
        self.snr_limit = snr_limit

    def run(self, sso_obs, orb, hval):
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return 0
        nights = len(np.unique(sso_obs[self.night_col][vis]))
        return nights


class ObsArcMetric(BaseMoMetric):
    """Calculate the difference between the first and last observation of an SSobject."""

    def __init__(self, snr_limit=None, **kwargs):
        super().__init__(**kwargs)
        self.snr_limit = snr_limit

    def run(self, sso_obs, orb, hval):
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return 0
        arc = sso_obs[self.mjd_col][vis].max() - sso_obs[self.mjd_col][vis].min()
        return arc


class DiscoveryMetric(BaseMoMetric):
    """Identify the discovery opportunities for an SSobject.

    Parameters
    ----------
    n_obs_per_night : int, optional
        Number of observations required within a single night. Default 2.
    t_min : float, optional
        Minimum time span between observations in a single night, in days.
        Default 5 minutes (5/60/24).
    t_max : float, optional
        Maximum time span between observations in a single night, in days.
        Default 90 minutes.
    n_nights_per_window : int, optional
        Number of nights required with observations, within the track window. Default 3.
    t_window : int, optional
        Number of nights included in the track window. Default 15.
    snr_limit : None or float, optional
        SNR limit to use for observations. If snr_limit is None, (default), then it uses
        the completeness calculation added to the 'vis' column (probabilistic visibility,
        based on 5-sigma limit). If snr_limit is not None, it uses this SNR value as a cutoff.
    metricName : str, optional
        The metric name to use.
        Default will be to construct Discovery_nObsPerNightxnNightsPerWindowintWindow.
    """

    def __init__(
        self,
        n_obs_per_night=2,
        t_min=5.0 / 60.0 / 24.0,
        t_max=90.0 / 60.0 / 24.0,
        n_nights_per_window=3,
        t_window=15,
        snr_limit=None,
        badval=None,
        **kwargs,
    ):
        # Define anything needed by the child metrics first.
        self.snr_limit = snr_limit
        self.child_metrics = {
            "N_Chances": DiscoveryNChancesMetric(self),
            "N_Obs": DiscoveryNObsMetric(self),
            "Time": DiscoveryTimeMetric(self),
            "Distance": DiscoveryDistanceMetric(self),
            "RADec": DiscoveryRadecMetric(self),
            "EcLonLat": DiscoveryEclonlatMetric(self),
        }
        if "metric_name" in kwargs:
            metric_name = kwargs.get("metric_name")
            del kwargs["metric_name"]
        else:
            metric_name = "Discovery_%.0fx%.0fin%.0f" % (
                n_obs_per_night,
                n_nights_per_window,
                t_window,
            )
        # Set up for inheriting from __init__.
        super().__init__(metric_name=metric_name, child_metrics=self.child_metrics, badval=badval, **kwargs)
        # Define anything needed for this metric.
        self.n_obs_per_night = n_obs_per_night
        self.t_min = t_min
        self.t_max = t_max
        self.n_nights_per_window = n_nights_per_window
        self.t_window = t_window

    def run(self, sso_obs, orb, hval):
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        # Identify discovery opportunities.
        #  Identify visits where the 'night' changes.
        vis_sort = np.argsort(sso_obs[self.mjd_col][vis])
        nights = sso_obs[self.night_col][vis][vis_sort]
        # print 'all nights', nights
        n = np.unique(nights)
        # Identify all the indexes where the night changes in value.
        n_idx = np.searchsorted(nights, n)
        # print 'nightchanges', nights[n_idx]
        # Count the number of observations per night (except last night)
        obs_per_night = (n_idx - np.roll(n_idx, 1))[1:]
        # Add the number of observations on the last night.
        obs_last_night = np.array([len(nights) - n_idx[-1]])
        obs_per_night = np.concatenate((obs_per_night, obs_last_night))
        # Find the nights with more than nObsPerNight.
        n_with_x_obs = n[np.where(obs_per_night >= self.n_obs_per_night)]
        n_idx_many = np.searchsorted(nights, n_with_x_obs)
        n_idx_many_end = np.searchsorted(nights, n_with_x_obs, side="right") - 1
        # Check that nObsPerNight observations are within tMin/tMax
        times_start = sso_obs[self.mjd_col][vis][vis_sort][n_idx_many]
        times_end = sso_obs[self.mjd_col][vis][vis_sort][n_idx_many_end]
        # Identify the nights with 'clearly good' observations.
        good = np.where(
            (times_end - times_start >= self.t_min) & (times_end - times_start <= self.t_max),
            1,
            0,
        )
        # Identify the nights where we need more investigation
        # (a subset of the visits may be within the interval).
        check = np.where(
            (good == 0)
            & (n_idx_many_end + 1 - n_idx_many > self.n_obs_per_night)
            & (times_end - times_start > self.t_max)
        )[0]
        for i, j, c in zip(vis_sort[n_idx_many][check], vis_sort[n_idx_many_end][check], check):
            t = sso_obs[self.mjd_col][vis][vis_sort][i : j + 1]
            dtimes = (np.roll(t, 1 - self.n_obs_per_night) - t)[:-1]
            tidx = np.where((dtimes >= self.t_min) & (dtimes <= self.t_max))[0]
            if len(tidx) > 0:
                good[c] = 1
        # 'good' provides mask for observations which could count as 'good to make tracklets'
        # against sso_obs[vis_sort][n_idx_many].  Now identify tracklets which can make tracks.
        good_idx = vis_sort[n_idx_many][good == 1]
        good_idx_ends = vis_sort[n_idx_many_end][good == 1]
        # print 'good tracklets', nights[good_idx]
        if len(good_idx) < self.n_nights_per_window:
            return self.badval
        delta_nights = (
            np.roll(sso_obs[self.night_col][vis][good_idx], 1 - self.n_nights_per_window)
            - sso_obs[self.night_col][vis][good_idx]
        )
        # Identify the index in sso_obs[vis][good_idx] (sorted by mjd) where the discovery opportunity starts.
        start_idxs = np.where((delta_nights >= 0) & (delta_nights <= self.t_window))[0]
        # Identify the index where the discovery opportunity ends.
        end_idxs = np.zeros(len(start_idxs), dtype="int")
        for i, sIdx in enumerate(start_idxs):
            in_window = np.where(
                sso_obs[self.night_col][vis][good_idx] - sso_obs[self.night_col][vis][good_idx][sIdx]
                <= self.t_window
            )[0]
            end_idxs[i] = np.array([in_window.max()])
        # Convert back to index based on sso_obs[vis] (sorted by expMJD).
        start_idxs = good_idx[start_idxs]
        end_idxs = good_idx_ends[end_idxs]
        return {
            "start": start_idxs[0:1],
            "end": end_idxs[0:1],
            "n_chances": len(start_idxs),
            # "trackletNights": sso_obs[self.night_col][vis][good_idx][0],
        }


class DiscoveryNChancesMetric(BaseChildMetric):
    """Calculate total number of discovery opportunities for an SSobject.

    Retirms total number of discovery opportunities.
    Child metric to be used with the Discovery Metric.
    """

    def __init__(
        self,
        parent_discovery_metric,
        # night_start=None,
        # night_end=None,
        badval=0,
        **kwargs,
    ):
        super().__init__(parent_discovery_metric, badval=badval, **kwargs)
        self.night_start = None  # night_start
        self.night_end = None  # night_end
        self.snr_limit = parent_discovery_metric.snr_limit
        # Update the metric name to use the night_start/night_end values, if an overriding name is not given.
        if "metric_name" not in kwargs:
            if self.night_start is not None:
                self.name = self.name + "_n%d" % (self.night_start)
            if self.night_end is not None:
                self.name = self.name + "_n%d" % (self.night_end)

    def run(self, sso_obs, orb, hval, metric_values):
        """Return the number of different discovery chances we had for each object/H combination."""
        return metric_values["n_chances"]
        """
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        if self.night_start is None and self.night_end is None:
            return len(metric_values["start"])
        # Otherwise, we have to sort out what night the discovery chances happened on.
        vis_sort = np.argsort(sso_obs[self.mjd_col][vis])
        nights = sso_obs[self.night_col][vis][vis_sort]
        start_nights = nights[metric_values["start"]]
        end_nights = nights[metric_values["end"]]
        if self.night_end is None and self.night_start is not None:
            valid = np.where(start_nights >= self.night_start)[0]
        elif self.night_start is None and self.night_end is not None:
            valid = np.where(end_nights <= self.night_end)[0]
        else:
            # And we only end up here if both were not None.
            valid = np.where(
                (start_nights >= self.night_start) & (end_nights <= self.night_end)
            )[0]
        return len(valid)
        """


class DiscoveryNObsMetric(BaseChildMetric):
    """Calculates the number of observations in the first discovery track of an SSobject."""

    def __init__(self, parent_discovery_metric, badval=0, **kwargs):
        super().__init__(parent_discovery_metric, badval=badval, **kwargs)
        # The number of the discovery chance to use.
        self.i = 0

    def run(self, sso_obs, orb, hval, metric_values):
        if self.i >= len(metric_values["start"]):
            return 0
        start_idx = metric_values["start"][self.i]
        end_idx = metric_values["end"][self.i]
        nobs = end_idx - start_idx
        return nobs


class DiscoveryTimeMetric(BaseChildMetric):
    """Returns the time of the first discovery track of an SSobject."""

    def __init__(self, parent_discovery_metric, t_start=None, badval=-999, **kwargs):
        super().__init__(parent_discovery_metric, badval=badval, **kwargs)
        self.i = 0
        self.t_start = t_start
        self.snr_limit = parent_discovery_metric.snr_limit

    def run(self, sso_obs, orb, hval, metric_values):
        if self.i >= len(metric_values["start"]):
            return self.badval
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        vis_sort = np.argsort(sso_obs[self.mjd_col][vis])
        times = sso_obs[self.mjd_col][vis][vis_sort]
        start_idx = metric_values["start"][self.i]
        t_disc = times[start_idx]
        if self.t_start is not None:
            t_disc = t_disc - self.t_start
        return t_disc


class DiscoveryDistanceMetric(BaseChildMetric):
    """Returns the distance of the first discovery track of an SSobject."""

    def __init__(self, parent_discovery_metric, distance_col="geo_dist", badval=-999, **kwargs):
        super().__init__(parent_discovery_metric, badval=badval, **kwargs)
        self.i = 0
        self.distance_col = distance_col
        self.snr_limit = parent_discovery_metric.snr_limit

    def run(self, sso_obs, orb, hval, metric_values):
        if self.i >= len(metric_values["start"]):
            return self.badval
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        vis_sort = np.argsort(sso_obs[self.mjd_col][vis])
        dists = sso_obs[self.distance_col][vis][vis_sort]
        start_idx = metric_values["start"][self.i]
        dist_disc = dists[start_idx]
        return dist_disc


class DiscoveryRadecMetric(BaseChildMetric):
    """Returns the RA/Dec of the first discovery track of an SSobject."""

    def __init__(self, parent_discovery_metric, badval=None, **kwargs):
        super().__init__(parent_discovery_metric, badval=badval, **kwargs)
        self.i = 0
        self.snr_limit = parent_discovery_metric.snr_limit
        self.metric_dtype = "object"

    def run(self, sso_obs, orb, hval, metric_values):
        if self.i >= len(metric_values["start"]):
            return self.badval
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        vis_sort = np.argsort(sso_obs[self.mjd_col][vis])
        ra = sso_obs[self.ra_col][vis][vis_sort]
        dec = sso_obs[self.dec_col][vis][vis_sort]
        start_idx = metric_values["start"][self.i]
        return (ra[start_idx], dec[start_idx])


class DiscoveryEclonlatMetric(BaseChildMetric):
    """Returns the ecliptic lon/lat and solar elong of the first discovery track of an SSobject."""

    def __init__(self, parent_discovery_metric, badval=None, **kwargs):
        super().__init__(parent_discovery_metric, badval=badval, **kwargs)
        self.i = 0
        self.snr_limit = parent_discovery_metric.snr_limit
        self.metric_dtype = "object"

    def run(self, sso_obs, orb, hval, metric_values):
        if self.i >= len(metric_values["start"]):
            return self.badval
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        vis_sort = np.argsort(sso_obs[self.mjd_col][vis])
        ec_lon = sso_obs["ec_lon"][vis][vis_sort]
        ec_lat = sso_obs["ec_lat"][vis][vis_sort]
        solar_elong = sso_obs["solar_elong"][vis][vis_sort]
        start_idx = metric_values["start"][self.i]
        return (ec_lon[start_idx], ec_lat[start_idx], solar_elong[start_idx])


class DiscoveryVelocityMetric(BaseChildMetric):
    """Returns the sky velocity of the first discovery track of an SSobject."""

    def __init__(self, parent_discovery_metric, badval=-999, **kwargs):
        super().__init__(parent_discovery_metric, badval=badval, **kwargs)
        self.i = 0
        self.snr_limit = parent_discovery_metric.snrLimit

    def run(self, sso_obs, orb, hval, metric_values):
        if self.i >= len(metric_values["start"]):
            return self.badval
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        vis_sort = np.argsort(sso_obs[self.mjd_col][vis])
        velocity = sso_obs["velocity"][vis][vis_sort]
        start_idx = metric_values["start"][self.i]
        return velocity[start_idx]


class ActivityOverTimeMetric(BaseMoMetric):
    """Count fraction of survey we could identify activity for an SSobject.

    Counts the time periods where we would have a chance to detect activity on
    a moving object.
    Splits observations into time periods set by 'window', then looks for observations within each window,
    and reports what fraction of the total windows receive 'nObs' visits.
    """

    def __init__(self, window, snr_limit=5, survey_years=10.0, metric_name=None, **kwargs):
        if metric_name is None:
            metric_name = "Chance of detecting activity lasting %.0f days" % (window)
        super().__init__(metric_name=metric_name, **kwargs)
        self.snr_limit = snr_limit
        self.window = window
        self.survey_years = survey_years
        self.window_bins = np.arange(0, self.survey_years * 365 + self.window / 2.0, self.window)
        self.n_windows = len(self.window_bins)
        self.units = "%.1f Day Windows" % (self.window)

    def run(self, sso_obs, orb, hval):
        # For cometary activity, expect activity at the same point in its orbit at the same time, mostly
        # For collisions, expect activity at random times
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        n, b = np.histogram(sso_obs[vis][self.night_col], bins=self.window_bins)
        activity_windows = np.where(n > 0)[0].size
        return activity_windows / float(self.n_windows)


class ActivityOverPeriodMetric(BaseMoMetric):
    """Count fraction of object period we could identify activity for an SSobject.

    Count the fraction of the orbit (when split into n_bins) that receive
    observations, in order to have a chance to detect activity.
    """

    def __init__(
        self,
        bin_size,
        snr_limit=5,
        q_col="q",
        e_col="e",
        a_col="a",
        t_peri_col="tPeri",
        anomaly_col="meanAnomaly",
        metric_name=None,
        **kwargs,
    ):
        """
        @ bin_size : size of orbit slice, in degrees.
        """
        if metric_name is None:
            metric_name = "Chance of detecting activity covering %.1f of the orbit" % (bin_size)
        super().__init__(metric_name=metric_name, **kwargs)
        self.q_col = q_col
        self.e_col = e_col
        self.a_col = a_col
        self.t_peri_col = t_peri_col
        self.anomaly_col = anomaly_col
        self.snr_limit = snr_limit
        self.bin_size = np.radians(bin_size)
        self.anomaly_bins = np.arange(0, 2 * np.pi, self.bin_size)
        self.anomaly_bins = np.concatenate([self.anomaly_bins, np.array([2 * np.pi])])
        self.n_bins = len(self.anomaly_bins) - 1
        self.units = "%.1f deg" % (np.degrees(self.bin_size))

    def run(self, sso_obs, orb, hval):
        # For cometary activity, expect activity at the same point in its orbit at the same time, mostly
        # For collisions, expect activity at random times
        if self.a_col in orb.keys():
            a = orb[self.a_col]
        elif self.q_col in orb.keys():
            a = orb[self.q_col] / (1 - orb[self.e_col])
        else:
            return self.badval

        period = np.power(a, 3.0 / 2.0) * 365.25  # days

        if self.anomaly_col in orb.keys():
            curranomaly = np.radians(
                orb[self.anomaly_col] + (sso_obs[self.mjd_col] - orb["epoch"]) / period * 360.0
            ) % (2 * np.pi)
        elif self.t_peri_col in orb.keys():
            curranomaly = ((sso_obs[self.mjd_col] - orb[self.t_peri_col]) / period) % (2 * np.pi)
        else:
            return self.badval

        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        n, b = np.histogram(curranomaly[vis], bins=self.anomaly_bins)
        activity_windows = np.where(n > 0)[0].size
        return activity_windows / float(self.n_bins)


class MagicDiscoveryMetric(BaseMoMetric):
    """Count the number of nights with discovery opportunities with very good software for an SSobject."""

    def __init__(self, n_obs=6, t_window=60, snr_limit=None, **kwargs):
        """
        @ n_obs = the total number of observations required for 'discovery'
        @ t_window = the timespan of the discovery window.
        @ snr_limit .. if snr_limit is None then uses 'completeness' calculation,
                   .. if snr_limit is not None, then uses this value as a cutoff.
        """
        super().__init__(**kwargs)
        self.snr_limit = snr_limit
        self.n_obs = n_obs
        self.t_window = t_window
        self.badval = 0

    def run(self, sso_obs, orb, hval):
        """SsoObs = Dataframe, orb=Dataframe, hval=single number."""
        # Calculate visibility for this orbit at this H.
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) < self.n_obs:
            return self.badval
        t_nights = np.sort(sso_obs[self.night_col])[vis]
        u_nights = np.unique(t_nights)
        indx_w = np.searchsorted(t_nights, u_nights, "left")
        indx_p = np.searchsorted(t_nights - self.t_window, u_nights, "left")
        n_disc = np.where(indx_p - indx_w >= self.n_obs)[0].size
        return n_disc


class HighVelocityMetric(BaseMoMetric):
    """Count number of times an SSobject appears trailed.

    Count the number of times an asteroid is observed with a velocity high enough to make it appear
    trailed by a factor of (psf_factor)*PSF - i.e. velocity >= psf_factor * seeing / visitExpTime.
    Simply counts the total number of observations with high velocity.
    """

    def __init__(self, psf_factor=2.0, snr_limit=None, velocity_col="velocity", **kwargs):
        """
        @ psf_factor = factor to multiply seeing/visitExpTime by
        (velocity(deg/day) >= 24*psf_factor*seeing(")/visitExptime(s))
        """
        super().__init__(**kwargs)
        self.velocity_col = velocity_col
        self.snr_limit = snr_limit
        self.psf_factor = psf_factor
        self.badval = 0

    def run(self, sso_obs, orb, hval):
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        high_velocity_obs = np.where(
            sso_obs[self.velocity_col][vis]
            >= (24.0 * self.psf_factor * sso_obs[self.seeing_col][vis] / sso_obs[self.exp_time_col][vis])
        )[0]
        return high_velocity_obs.size


class HighVelocityNightsMetric(BaseMoMetric):
    """Count the number of discovery opportunities (via trailing) for an SSobject.

    Determine the first time an asteroid is observed is observed with a velocity high enough to make
    it appear trailed by a factor of psf_factor*PSF with n_obs_per_night observations within a given night.

    Parameters
    ----------
    psf_factor: float, optional
        Object velocity (deg/day) must be >= 24 * psf_factor * seeingGeom (") / visitExpTime (s).
        Default is 2 (i.e. object trailed over 2 psf's).
    n_obs_per_night: int, optional
        Number of observations per night required. Default 2.
    snr_limit: float or None
        If snr_limit is set as a float, then requires object to be above snr_limit SNR in the image.
        If snr_limit is None, this uses the probabilistic 'visibility' calculated by the vis stacker,
        which means SNR ~ 5.   Default is None.
    velocity_col: str, optional
        Name of the velocity column in the obs file. Default 'velocity'. (note this is deg/day).

    Returns
    -------
    float
        The time of the first detection where the conditions are satisifed.
    """

    def __init__(self, psf_factor=2.0, n_obs_per_night=2, snr_limit=None, velocity_col="velocity", **kwargs):
        super().__init__(**kwargs)
        self.velocity_col = velocity_col
        self.snr_limit = snr_limit
        self.psf_factor = psf_factor
        self.n_obs_per_night = n_obs_per_night

    def run(self, sso_obs, orb, hval):
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return self.badval
        high_velocity_obs = np.where(
            sso_obs[self.velocity_col][vis]
            >= (24.0 * self.psf_factor * sso_obs[self.seeing_col][vis] / sso_obs[self.exp_time_col][vis])
        )[0]
        if len(high_velocity_obs) == 0:
            return 0
        nights = sso_obs[self.night_col][vis][high_velocity_obs]
        n = np.unique(nights)
        n_idx = np.searchsorted(nights, n)
        # Count the number of observations per night (except last night)
        obs_per_night = (n_idx - np.roll(n_idx, 1))[1:]
        # Add the number of observations on the last night.
        obs_last_night = np.array([len(nights) - n_idx[-1]])
        obs_per_night = np.concatenate((obs_per_night, obs_last_night))
        # Find the nights with at least nObsPerNight visits
        # (this is already looking at only high velocity observations).
        n_with_x_obs = n[np.where(obs_per_night >= self.n_obs_per_night)]
        if len(n_with_x_obs) > 0:
            found = sso_obs[np.where(sso_obs[self.night_col] == n_with_x_obs[0])][self.mjd_col][0]
        else:
            found = self.badval
        return found


class LightcurveInversionAsteroidMetric(BaseMoMetric):
    """
    This metric is generally applicable to NEOs and MBAs - inner solar system objects.

    Determine if the cumulative sum of observations of a target are enough to enable lightcurve
    inversion for shape modeling. For this to be true, multiple conditions need to be
    satisfied:

    1) The SNR-weighted number of observations (each observation is weighted by its SNR, up to a max of 100)
    must be larger than the threshhold weight_det (default 50)
    2) Ecliptic longitudinal coverage needs to be at least 90 degrees, and the absolute deviation
    needs to be at least 1/8th the longitudinal coverage.
    3) The phase angle coverage needs to span at least 5 degrees.

    For evaluation of condition 2, the median ecliptic longitude is subtracted from all longitudes,
    and the modulo 360 of those values is taken. This ensures that the wrap around 360 is handled
    correctly.

    For more information on the above conditions, please see
    https://docs.google.com/document/d/1GAriM7trpTS08uanjUF7PyKALB2JBTjVT7Y6R30i0-8/edit?usp=sharing
    Contributed by Steve Chesley, Wes Fraser, Josef Durech, and the inner solar system working group.

    Parameters
    ----------
    weight_det: float, optional
        The SNR-weighted number of detections required (per bandpass in any ONE of the filters in filterlist).
        Default 50.
    snr_limit: float or None, optional
        If snr_limit is set as a float, then requires object to be above snr_limit SNR in the image.
        If snr_limit is None, this uses the probabilistic 'visibility' calculated by the vis stacker,
        which means SNR ~ 5.   Default is None.
    snr_max: float, optional
        Maximum value toward the SNR-weighting to consider. Default 100.
    filterlist: list of str, optional
        The filters which the lightcurve inversion could be based on. Requirements must be met in one of
        these filters.

    Returns
    -------
    int
        0 (could not perform lightcurve inversion) or 1 (could)
    """

    def __init__(
        self, weight_det=50, snr_limit=None, snr_max=100, filterlist=("u", "g", "r", "i", "z", "y"), **kwargs
    ):
        super().__init__(**kwargs)
        self.snr_limit = snr_limit
        self.snr_max = snr_max
        self.weight_det = weight_det
        self.filterlist = filterlist

    def run(self, sso_obs, orb, hval):
        # Calculate the clipped SNR - ranges from snrLimit / SNR+vis to snrMax.
        clip_snr = np.minimum(sso_obs[self.snr_col], self.snr_max)
        if self.snr_limit is not None:
            clip_snr = np.where(sso_obs[self.snr_col] <= self.snr_limit, 0, clip_snr)
        else:
            clip_snr = np.where(sso_obs[self.vis_col] == 0, 0, clip_snr)
        if len(np.where(clip_snr > 0)[0]) == 0:
            return 0
        # Check each filter in filterlist:
        # stop as soon as find a filter that matches requirements.
        inversion_possible = 0
        for f in self.filterlist:
            # Is the SNR-weight sum of observations in this filter high enough?
            match = np.where(sso_obs[self.filter_col] == f)
            snr_sum = np.sum(clip_snr[match]) / self.snr_max
            if snr_sum < self.weight_det:
                # Do not have enough SNR-weighted observations, so skip on to the next filter.
                continue
            # Is the ecliptic longitude coverage for the visible observations sufficient?
            # Is the phase coverage sufficient?
            vis = np.where(clip_snr[match] > 0)
            ec_l = sso_obs["ecLon"][match][vis]
            phase_angle = sso_obs["phase"][match][vis]
            # Calculate the absolute deviation and range of ecliptic longitude.
            ec_l_centred = (ec_l - np.median(ec_l)) % 360.0
            a_dev = np.sum(np.abs(ec_l_centred - np.mean(ec_l_centred))) / len(ec_l_centred)
            d_l = np.max(ec_l) - np.min(ec_l)
            # Calculate the range of the phase angle
            dp = np.max(phase_angle) - np.min(phase_angle)
            # Metric requirement is that d_l >= 90 deg, absolute deviation is greater than d_l/8
            # and then that the phase coverage is more than 5 degrees.
            # Stop as soon as find a case where this is true.
            if d_l >= 90.0 and a_dev >= d_l / 8 and dp >= 5:
                inversion_possible += 1
                break
        return inversion_possible


class ColorAsteroidMetric(BaseMoMetric):
    """
    This metric is appropriate for MBAs and NEOs, and other inner solar system objects.

    The metric evaluates if the SNR-weighted number of observations are enough to
    determine an approximate lightcurve and phase function -- and from this,
    then a color for the asteroid can be determined.
    The assumption is that you must fit the lightcurve/phase function in each bandpass,
    and could do this well-enough if you have at least weight_det SNR-weighted observations
    in the bandpass.
    e.g. to find a g-r color, you must have 10 (SNR-weighted) obs in g and 10 in r.

    For more details, see
    https://docs.google.com/document/d/1GAriM7trpTS08uanjUF7PyKALB2JBTjVT7Y6R30i0-8/edit?usp=sharing
    Contributed by Wes Fraser, Steven Chesley & the inner solar system working group.

    Parameters
    ----------
    weight_det: float, optional
        The SNR-weighted number of detections required (per bandpass in any ONE of the filters in filterlist).
        Default 10.
    snr_limit: float or None, optional
        If snr_limit is set as a float, then requires object to be above snr_limit SNR in the image.
        If snr_limit is None, this uses the probabilistic 'visibility' calculated by the vis stacker,
        which means SNR ~ 5.   Default is None.
    snr_max: float, optional
        Maximum value toward the SNR-weighting to consider. Default 20.

    Returns
    -------
    int
        An integer 'flag' that indicates whether the mean magnitude (and thus a color) was determined in:
        0 = no bands
        1 = g and (r or i) and (z or y). i.e. obtain colors g-r or g-i PLUS g-z or g-y
        2 = Any 4 different filters (from grizy). i.e. colors = g-r, r-i, i-z, OR r-i, i-z, z-y..
        3 = All 5 from grizy. i.e. colors g-r, r-i, i-z, z-y.
        4 = All 6 filters (ugrizy) -- best possible! add u-g.
    """

    def __init__(self, weight_det=10, snr_max=20, snr_limit=None, **kwargs):
        super().__init__(**kwargs)
        self.weight_det = weight_det
        self.snr_limit = snr_limit
        self.snr_max = snr_max
        self.filterlist = ("u", "g", "r", "i", "z", "y")

    def run(self, sso_obs, orb, hval):
        clip_snr = np.minimum(sso_obs[self.snr_col], self.snr_max)
        if self.snr_limit is not None:
            clip_snr = np.where(sso_obs[self.snr_col] <= self.snr_limit, 0, clip_snr)
        else:
            clip_snr = np.where(sso_obs[self.vis_col] == 0, 0, clip_snr)
        if len(np.where(clip_snr > 0)[0]) == 0:
            return self.badval

        # Evaluate SNR-weighted number of observations in each filter.
        filter_weight = {}
        for f in self.filterlist:
            match = np.where(sso_obs[self.filter_col] == f)
            snrweight = np.sum(clip_snr[match]) / self.snr_max
            # If the snrweight exceeds the weightDet, add it to the dictionary.
            if snrweight > self.weight_det:
                filter_weight[f] = snrweight

        # Now assign a flag:
        # 0 = no bands
        # 1 = g and (r or i) and (z or y). i.e. obtain colors g-r or g-i PLUS g-z or g-y
        # 2 = Any 4 different filters (from grizy). i.e. colors = g-r, r-i, i-z, OR r-i, i-z, z-y..
        # 3 = All 5 from grizy. i.e. colors g-r, r-i, i-z, z-y.
        # 4 = All 6 filters (ugrizy) -- best possible! add u-g.
        all_six = set(self.filterlist)
        good_five = set(["g", "r", "i", "z", "y"])

        if len(filter_weight) == 0:  # this lets us stop evaluating here if possible.
            flag = 0
        elif all_six.intersection(filter_weight) == all_six:
            flag = 4
        elif good_five.intersection(filter_weight) == good_five:
            flag = 3
        elif len(good_five.intersection(filter_weight)) == 4:
            flag = 2
        elif "g" in filter_weight:
            # Have 'g' - do we have (r or i) and (z or y)
            if ("r" in filter_weight or "i" in filter_weight) and (
                "z" in filter_weight or "y" in filter_weight
            ):
                flag = 1
            else:
                flag = 0
        else:
            flag = 0

        return flag


class LightcurveColorOuterMetric(BaseMoMetric):
    """
    This metric is appropriate for outer solar system objects, such as TNOs and SDOs.

    This metric evaluates whether the number of observations is sufficient to fit a lightcurve
    in a primary and secondary bandpass. The primary bandpass requires more observations than
    the secondary. Essentially, it's a complete lightcurve in one or both bandpasses, with at
    least a semi-complete lightcurve in the secondary band.

    The lightcurve/color can be calculated with any two of the bandpasses in filterlist.
    Contributed by Wes Fraser.

    Parameters
    ----------
    snr_limit: float or None, optional
        If snr_limit is set as a float, then requires object to be above snr_limit SNR in the image.
        If snr_limit is None, this uses the probabilistic 'visibility' calculated by the vis stacker,
        which means SNR ~ 5.   Default is None.
    num_req: int, optional
        Number of observations required for a lightcurve fitting. Default 30.
    num_sec_filt: int, optional
        Number of observations required in a secondary band for color only. Default 20.
    filterlist: list of str, optional
        Filters that the primary/secondary measurements can be in.

    Returns
    -------
    int
        A flag that indicates whether a color/lightcurve was generated in:
        0 = no lightcurve (although may have had 'color' in one or more band)
        1 = a lightcurve in a single filter (but no additional color information)
        2+ = lightcurves in more than one filter  (or lightcurve + color)
        e.g. lightcurve in 2 bands, with additional color information in another = 3.
    """

    def __init__(
        self, snr_limit=None, num_req=30, num_sec_filt=20, filterlist=("u", "g", "r", "i", "z", "y"), **kwargs
    ):
        super().__init__(**kwargs)
        self.snr_limit = snr_limit
        self.num_req = num_req
        self.num_sec_filt = num_sec_filt
        self.filterlist = filterlist

    def run(self, sso_obs, orb, hval):
        vis = _set_vis(sso_obs, self.snr_limit, self.snr_col, self.vis_col)
        if len(vis) == 0:
            return 0

        lightcurves = set()
        colors = set()
        for f in self.filterlist:
            nmatch = np.where(sso_obs[vis][self.filter_col] == f)[0]
            if len(nmatch) >= self.num_req:
                lightcurves.add(f)
            if len(nmatch) >= self.num_sec_filt:
                colors.add(f)

        # Set the flags - first the number of filters with lightcurves.
        flag = len(lightcurves)
        # And check if there were extra filters which had enough for a color
        # but not enough for a full lightcurve.
        if len(colors.difference(lightcurves)) > 0:
            # If there was no lightcurve available to match against:
            if len(lightcurves) == 0:
                flag = 0
            else:
                # We had a lightcurve and now can add a color.
                flag += 1
        return flag


class InstantaneousColorMetric(BaseMoMetric):
    """Identify SSobjects which could have observations suitable to determine colors.

    Generally, this is not the mode LSST would work in - the lightcurves of the objects
    mean that the time interval would have to be quite short.

    This is roughly defined as objects which have more than n_pairs pairs of observations
    with SNR greater than snr_limit, in bands bandOne and bandTwo, within n_hours.

    Parameters
    ----------
    n_pairs: int, optional
        The number of pairs of observations (in each band) that must be within n_hours
        Default 1
    snr_limit: float, optional
        The SNR limit for the observations. Default 10.
    n_hours: float, optional
        The time interval between observations in the two bandpasses (hours). Default 0.5 hours.
    b_one: str, optional
        The first bandpass for the color. Default 'g'.
    b_two: str, optional
        The second bandpass for the color. Default 'r'.

    Returns
    -------
    int
        0 (no color possible under these constraints) or 1 (color possible).
    """

    def __init__(self, n_pairs=1, snr_limit=10, n_hours=0.5, b_one="g", b_two="r", **kwargs):
        super().__init__(**kwargs)
        self.n_pairs = n_pairs
        self.snr_limit = snr_limit
        self.n_hours = n_hours
        self.b_one = b_one
        self.b_two = b_two
        self.badval = -666

    def run(self, sso_obs, orb, hval):
        vis = np.where(sso_obs[self.snr_col] >= self.snr_limit)[0]
        if len(vis) < self.n_pairs * 2:
            return 0
        b_one_obs = np.where(sso_obs[self.filter_col][vis] == self.b_one)[0]
        b_two_obs = np.where(sso_obs[self.filter_col][vis] == self.b_two)[0]
        timesb_one = sso_obs[self.mjd_col][vis][b_one_obs]
        timesb_two = sso_obs[self.mjd_col][vis][b_two_obs]
        if len(timesb_one) == 0 or len(timesb_two) == 0:
            return 0
        d_time = self.n_hours / 24.0
        # Calculate the time between the closest pairs of observations.
        in_order = np.searchsorted(timesb_one, timesb_two, "right")
        in_order = np.where(in_order - 1 > 0, in_order - 1, 0)
        dt_pairs = timesb_two - timesb_one[in_order]
        if len(np.where(dt_pairs < d_time)[0]) >= self.n_pairs:
            found = 1
        else:
            found = 0
        return found


class PeakVMagMetric(BaseMoMetric):
    """Pull out the peak V magnitude of all observations of the SSobject."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def run(self, sso_obs, orb, hval):
        peak_vmag = np.min(sso_obs[self.app_mag_v_col])
        return peak_vmag


class KnownObjectsMetric(BaseMoMetric):
    """Identify SSobjects which could be classified as 'previously known' based on their peak V magnitude.
    This is most appropriate for NEO surveys, where most of the sky has been covered so the exact location
    (beyond being in the visible sky) is not as important.

    Default parameters tuned to match NEO survey capabilities.
    Returns the time at which each first reached that threshold V magnitude.
    The default values are calibrated using the NEOs larger than 140m discovered in the last 20 years
    and assuming a 30% completeness in 2017.

    Parameters
    -----------
    elong_thresh : float, optional
        The cutoff in solar elongation to consider an object 'visible'. Default 100 deg.
    v_mag_thresh1 : float, optional
        The magnitude threshold for previously known objects. Default 20.0.
    eff1 : float, optional
        The likelihood of actually achieving each individual input observation.
        If the input observations include one observation per day, an 'eff' value of 0.3 would
        mean that (on average) only one third of these observations would be achieved.
        This is similar to the level for LSST, which can cover the visible sky every 3-4 days.
        Default 0.1
    t_switch1 : float, optional
        The (MJD) time to switch between v_mag_thresh1 + eff1 to v_mag_thresh2 + eff2, e.g.
        the end of the first period.
        Default 53371 (2005).
    v_mag_thresh2 : float, optional
        The magnitude threshhold for previously known objects. Default 22.0.
        This is based on assuming PS and other surveys will be efficient down to V=22.
    eff2 : float, optional
        The efficiency of observations during the second period of time. Default 0.1
    t_switch2 : float, optional
        The (MJD) time to switch between v_mag_thresh2 + eff2 to v_mag_thresh3 + eff3.
        Default 57023 (2015).
    v_mag_thresh3 : float, optional
        The magnitude threshold during the third period. Default 22.0, based on PS1 + Catalina.
    eff3 : float, optional
        The efficiency of observations during the third period. Default 0.1
    t_switch3 : float, optional
        The (MJD) time to switch between v_mag_thresh3 + eff3 to v_mag_thresh4 + eff4.
        Default 59580 (2022).
    v_mag_thresh4 : float, optional
        The magnitude threshhold during the fourth (last) period. Default 22.0, based on PS1 + Catalina.
    eff4 : float, optional
        The efficiency of observations during the fourth (last) period. Default 0.2
    """

    def __init__(
        self,
        elong_thresh=100.0,
        v_mag_thresh1=20.0,
        eff1=0.1,
        t_switch1=53371,  # XXX--maybe swap to survey_start_mjd and then delta_t's
        v_mag_thresh2=21.5,
        eff2=0.1,
        t_switch2=57023,
        v_mag_thresh3=22.0,
        eff3=0.1,
        t_switch3=59580,
        v_mag_thresh4=22.0,
        eff4=0.2,
        elong_col="Elongation",
        mjd_col="MJD(UTC)",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.elong_thresh = elong_thresh
        self.elong_col = elong_col
        self.v_mag_thresh1 = v_mag_thresh1
        self.eff1 = eff1
        self.t_switch1 = t_switch1
        self.v_mag_thresh2 = v_mag_thresh2
        self.eff2 = eff2
        self.t_switch2 = t_switch2
        self.v_mag_thresh3 = v_mag_thresh3
        self.eff3 = eff3
        self.t_switch3 = t_switch3
        self.v_mag_thresh4 = v_mag_thresh4
        self.eff4 = eff4
        self.mjd_col = mjd_col
        self.badval = int(t_switch3) + 365 * 1000

    def _pick_obs(self, potential_obs_times, eff):
        # From a set of potential observations, apply an efficiency
        # And return the minimum time (if any)
        rand_pick = np.random.rand(len(potential_obs_times))
        picked = np.where(rand_pick <= eff)[0]
        if len(picked) > 0:
            disc_time = potential_obs_times[picked].min()
        else:
            disc_time = None
        return disc_time

    def run(self, sso_obs, orb, hval):
        visible = np.where(sso_obs[self.elong_col] >= self.elong_thresh, 1, 0)
        discovery_time = None
        # Look for discovery in any of the three periods.
        # First period.
        obs1 = np.where((sso_obs[self.mjd_col] < self.t_switch1) & visible)[0]
        over_peak = np.where(sso_obs[self.app_mag_v_col][obs1] <= self.v_mag_thresh1)[0]
        if len(over_peak) > 0:
            discovery_time = self._pick_obs(sso_obs[self.mjd_col][obs1][over_peak], self.eff1)
        # Second period.
        if discovery_time is None:
            obs2 = np.where(
                (sso_obs[self.mjd_col] >= self.t_switch1) & (sso_obs[self.mjd_col] < self.t_switch2) & visible
            )[0]
            over_peak = np.where(sso_obs[self.app_mag_v_col][obs2] <= self.v_mag_thresh2)[0]
            if len(over_peak) > 0:
                discovery_time = self._pick_obs(sso_obs[self.mjd_col][obs2][over_peak], self.eff2)
        # Third period.
        if discovery_time is None:
            obs3 = np.where(
                (sso_obs[self.mjd_col] >= self.t_switch2) & (sso_obs[self.mjd_col] < self.t_switch3) & visible
            )[0]
            over_peak = np.where(sso_obs[self.app_mag_v_col][obs3] <= self.v_mag_thresh3)[0]
            if len(over_peak) > 0:
                discovery_time = self._pick_obs(sso_obs[self.mjd_col][obs3][over_peak], self.eff3)
        # Fourth period.
        if discovery_time is None:
            obs4 = np.where((sso_obs[self.mjd_col] >= self.t_switch3) & visible)[0]
            over_peak = np.where(sso_obs[self.app_mag_v_col][obs4] <= self.v_mag_thresh4)[0]
            if len(over_peak) > 0:
                discovery_time = self._pick_obs(sso_obs[self.mjd_col][obs4][over_peak], self.eff4)
        if discovery_time is None:
            discovery_time = self.badval
        return discovery_time



__all__ = (
    "FootprintFractionMetric",
    "FOArea",
    "FONv",
    "IdentityMetric",
    "NormalizeMetric",
    "ZeropointMetric",
    "TotalPowerMetric",
    "StaticProbesFoMEmulatorMetricSimple",
)

import warnings

import healpy as hp
import numpy as np
from scipy import interpolate

from .base_metric import BaseMetric

# Metrics which are primarily intended to be used as summary statistics.


class FootprintFractionMetric(BaseMetric):
    """Calculate fraction of a desired footprint got covered.
    Helpful to check if everything was covered in first year

    Parameters
    ----------
    footprint : `np.ndarray`, (N,)
        The HEALpix footprint to compare to.
        Nside of the footprint should match nside of the slicer.
    n_min : `int`
        The number of visits to require to consider an area covered
    """

    def __init__(self, footprint=None, n_min=1, **kwargs):
        super().__init__(**kwargs)
        self.footprint = footprint
        self.nside = hp.npix2nside(footprint.size)
        self.npix = np.where(self.footprint > 0)[0].size
        # get whole array passed
        self.mask_val = 0
        self.n_min = n_min

    def run(self, data_slice, slice_point=None):
        overlap = np.where((self.footprint > 0) & (data_slice["metricdata"] >= self.n_min))[0]
        result = overlap.size / self.npix
        return result


class FONv(BaseMetric):
    """Given asky area, what is the minimum and median NVISITS obtained over
    that area?
    (chooses the portion of the sky with the highest number of visits first).

    Parameters
    ----------
    col : `str` or `list` of `strs`, optional
        Name of the column in the numpy recarray passed to the summary metric.
    asky : `float`, optional
        Area of the sky to base the evaluation of number of visits over.
    nside : `int`, optional
        Nside parameter from healpix slicer, used to set the physical
        relationship between on-sky area and number of healpixels.
    n_visit : `int`, optional
        Number of visits to use as the benchmark value, if choosing to return
        a normalized n_visit value.
    norm : `bool`, optional
        Normalize the returned "n_visit" (min / median) values by n_visit,
        if true.
    metric_name : `str`, optional
        Name of the summary metric. Default FONv.
    """

    def __init__(self, col="metricdata", asky=18000.0, nside=128, n_visit=825, norm=False, **kwargs):
        """asky = square degrees"""
        super().__init__(col=col, **kwargs)
        self.nvisit = n_visit
        self.nside = nside
        # Determine how many healpixels are included in asky sq deg.
        self.asky = asky
        self.scale = hp.nside2pixarea(self.nside, degrees=True)
        self.npix__asky = int(np.ceil(self.asky / self.scale))
        self.norm = norm

    def run(self, data_slice, slice_point=None):
        result = np.empty(2, dtype=[("name", np.str_, 20), ("value", float)])
        result["name"][0] = "MedianNvis"
        result["name"][1] = "MinNvis"
        # If there is not even as much data as needed to cover Asky:
        if len(data_slice) < self.npix__asky:
            result["value"][0] = self.badval
            result["value"][1] = self.badval
            return result
        # Otherwise, calculate median and mean Nvis:
        name = data_slice.dtype.names[0]
        nvis_sorted = np.sort(data_slice[name])
        # Find the Asky's worth of healpixels with the largest # of visits.
        nvis__asky = nvis_sorted[-self.npix__asky :]
        result["value"][0] = np.median(nvis__asky)
        result["value"][1] = np.min(nvis__asky)
        if self.norm:
            result["value"] /= float(self.nvisit)
        return result


class FOArea(BaseMetric):
    """Given an n_visit threshold, how much AREA receives at least that many
    visits?

    Parameters
    ----------
    col : `str` or `list` of `strs`, optional
        Name of the column in the numpy recarray passed to the summary metric.
    n_visit : `int`, optional
        Number of visits to use as the minimum required --
        metric calculated area that has this many visits.
    asky : `float`, optional
        Area to use as the benchmark area value,
        if choosing to return a normalized Area value.
    nside : `int`, optional
        Nside parameter from healpix slicer, used to set the physical
        relationship between on-sky area and number of healpixels.
    norm : `bool`, optional
        If true, normalize the returned area value by asky.
    """

    def __init__(
        self,
        col="metricdata",
        n_visit=825,
        asky=18000.0,
        nside=128,
        norm=False,
        **kwargs,
    ):
        """asky = square degrees"""
        super().__init__(col=col, **kwargs)
        self.nvisit = n_visit
        self.nside = nside
        self.asky = asky
        self.scale = hp.nside2pixarea(self.nside, degrees=True)
        self.norm = norm

    def run(self, data_slice, slice_point=None):
        name = data_slice.dtype.names[0]
        nvis_sorted = np.sort(data_slice[name])
        # Identify the healpixels with more than Nvisits.
        nvis_min = nvis_sorted[np.where(nvis_sorted >= self.nvisit)]
        if len(nvis_min) == 0:
            result = self.badval
        else:
            result = nvis_min.size * self.scale
            if self.norm:
                result /= float(self.asky)
        return result


class IdentityMetric(BaseMetric):
    """
    Return the metric value itself .. this is primarily useful as a summary statistic for UniSlicer metrics.
    """

    def run(self, data_slice, slice_point=None):
        if len(data_slice[self.colname]) == 1:
            result = data_slice[self.colname][0]
        else:
            result = data_slice[self.colname]
        return result


class NormalizeMetric(BaseMetric):
    """
    Return a metric values divided by 'norm_val'. Useful for turning summary statistics into fractions.
    """

    def __init__(self, col="metricdata", norm_val=1, **kwargs):
        super(NormalizeMetric, self).__init__(col=col, **kwargs)
        self.norm_val = float(norm_val)

    def run(self, data_slice, slice_point=None):
        result = data_slice[self.colname] / self.norm_val
        if len(result) == 1:
            return result[0]
        else:
            return result


class ZeropointMetric(BaseMetric):
    """
    Return a metric values with the addition of 'zp'. Useful for altering the zeropoint for summary statistics.
    """

    def __init__(self, col="metricdata", zp=0, **kwargs):
        super(ZeropointMetric, self).__init__(col=col, **kwargs)
        self.zp = zp

    def run(self, data_slice, slice_point=None):
        result = data_slice[self.colname] + self.zp
        if len(result) == 1:
            return result[0]
        else:
            return result


class TotalPowerMetric(BaseMetric):
    """
    Calculate the total power in the angular power spectrum between lmin/lmax.
    """

    def __init__(
        self, col="metricdata", lmin=100.0, lmax=300.0, remove_dipole=True, mask_val=np.nan, **kwargs
    ):
        self.lmin = lmin
        self.lmax = lmax
        self.remove_dipole = remove_dipole
        super(TotalPowerMetric, self).__init__(col=col, mask_val=mask_val, **kwargs)

    def run(self, data_slice, slice_point=None):
        # Calculate the power spectrum.
        if self.remove_dipole:
            cl = hp.anafast(hp.remove_dipole(data_slice[self.colname], verbose=False))
        else:
            cl = hp.anafast(data_slice[self.colname])
        ell = np.arange(np.size(cl))
        condition = np.where((ell <= self.lmax) & (ell >= self.lmin))[0]
        totalpower = np.sum(cl[condition] * (2 * ell[condition] + 1))
        return totalpower


class StaticProbesFoMEmulatorMetricSimple(BaseMetric):
    """This calculates the Figure of Merit for the combined
    static probes (3x2pt, i.e., Weak Lensing, LSS, Clustering).
    This FoM is purely statistical and does not factor in systematics.

    This version of the emulator was used to generate the results in
    https://ui.adsabs.harvard.edu/abs/2018arXiv181200515L/abstract

    A newer version is being created. This version has been renamed
    Simple in anticipation of the newer, more sophisticated metric
    replacing it.

    Note that this is truly a summary metric and should be run on the output of
    Exgalm5_with_cuts.
    """

    def __init__(self, nside=128, year=10, col=None, **kwargs):
        """
        Args:
            nside (int): healpix resolution
            year (int): year of the FoM emulated values,
                can be one of [1, 3, 6, 10]
            col (str): column name of metric data.
        """
        self.nside = nside
        super().__init__(col=col, **kwargs)
        if col is None:
            self.col = "metricdata"
        self.year = year

    def run(self, data_slice, slice_point=None):
        """
        Args:
            data_slice (ndarray): Values passed to metric by the slicer,
                which the metric will use to calculate metric values
                at each slice_point.
            slice_point (Dict): Dictionary of slice_point metadata passed
                to each metric.
        Returns:
             float: Interpolated static-probe statistical Figure-of-Merit.
        Raises:
             ValueError: If year is not one of the 4 for which a FoM is calculated
        """
        # Chop off any outliers
        good_pix = np.where(data_slice[self.col] > 0)[0]

        # Calculate area and med depth from
        area = hp.nside2pixarea(self.nside, degrees=True) * np.size(good_pix)
        median_depth = np.median(data_slice[self.col][good_pix])

        # FoM is calculated at the following values
        if self.year == 1:
            areas = [7500, 13000, 16000]
            depths = [24.9, 25.2, 25.5]
            fom_arr = [
                [1.212257e02, 1.462689e02, 1.744913e02],
                [1.930906e02, 2.365094e02, 2.849131e02],
                [2.316956e02, 2.851547e02, 3.445717e02],
            ]
        elif self.year == 3:
            areas = [10000, 15000, 20000]
            depths = [25.5, 25.8, 26.1]
            fom_arr = [
                [1.710645e02, 2.246047e02, 2.431472e02],
                [2.445209e02, 3.250737e02, 3.516395e02],
                [3.173144e02, 4.249317e02, 4.595133e02],
            ]

        elif self.year == 6:
            areas = [10000, 15000, 2000]
            depths = [25.9, 26.1, 26.3]
            fom_arr = [
                [2.346060e02, 2.414678e02, 2.852043e02],
                [3.402318e02, 3.493120e02, 4.148814e02],
                [4.452766e02, 4.565497e02, 5.436992e02],
            ]

        elif self.year == 10:
            areas = [10000, 15000, 20000]
            depths = [26.3, 26.5, 26.7]
            fom_arr = [
                [2.887266e02, 2.953230e02, 3.361616e02],
                [4.200093e02, 4.292111e02, 4.905306e02],
                [5.504419e02, 5.624697e02, 6.441837e02],
            ]
        else:
            warnings.warn("FoMEmulator is not defined for this year")
            return self.badval

        # Interpolate FoM to the actual values for this sim
        areas = [[i] * 3 for i in areas]
        depths = [depths] * 3
        f = interpolate.interp2d(areas, depths, fom_arr, bounds_error=False)
        fom = f(area, median_depth)[0]
        return fom



__all__ = ("KuiperMetric",)

import numpy as np

from .base_metric import BaseMetric


class KuiperMetric(BaseMetric):
    """Find the Kuiper V statistic for a distribution, useful for angles.

    Value of 0 means perfecty uniform, 1 means delta function
    """

    def run(self, data_slice, slice_point=None):
        """"""
        # Assume input in degrees
        values = np.sort(data_slice[self.colname] % 360)

        dist_1 = (np.arange(values.size) + 1) / values.size
        uniform = values / (360.0)

        d_plus = np.max(uniform - dist_1)
        d_minus = np.max(dist_1 - uniform)
        result = d_plus + d_minus

        return result



__all__ = (
    "GapsMetric",
    "TgapsMetric",
    "TgapsPercentMetric",
    "NightgapsMetric",
    "NVisitsPerNightMetric",
    "MaxGapMetric",
    "NightTimespanMetric",
)

import numpy as np

from .base_metric import BaseMetric


class GapsMetric(BaseMetric):
    """Compute the number of times a gaps of a specified timescale (+/- 1/2 timescale) is sampled.

    Parameters
    ----------
    times_col : `str` (observationStartMJD)
        The column name for the exposure times.  Values assumed to be in days.
    time_scale : `float` (2)
        Time scale to see how well it is sampled (hours). For example, the default of 2 hours means
        observations spaced anywhere between 1 and 3 hours will count as a sample.
    """

    def __init__(
        self,
        times_col="observationStartMJD",
        time_scale=2.0,
        units="N",
        **kwargs,
    ):
        self.times_col = times_col
        # Convert to days and divide by two so we bin at the Nyquist frequency
        self.bin_size = (time_scale / 24.0) / 2.0
        super().__init__(col=[self.times_col], metric_dtype="float", units=units, **kwargs)

    def run(self, data_slice, slice_point=None):
        if data_slice.size < 2:
            return self.badval
        times = np.sort(data_slice[self.times_col])
        bins = np.arange(times.min() - self.bin_size, times.max() + self.bin_size, self.bin_size)
        vals, _be = np.histogram(times, bins)
        # Bins are at 1/2 desired timescale, so compare bin 0 to bin 2 to see
        # if the timescale has been sampled.
        mult = vals[2:] * vals[0:-2]
        result = np.size(np.where(mult > 0)[0])

        return result


class TgapsMetric(BaseMetric):
    """Histogram the times of the gaps between observations.


    Measure the gaps between observations.  By default, only gaps
    between neighboring visits are computed.  If all_gaps is set to true, all gaps are
    computed (i.e., if there are observations at 10, 20, 30 and 40 the default will
    return a histogram of [10,10,10] while all_gaps returns a histogram of [10,10,10,20,20,30])

    Parameters
    ----------
    times_col : `str`, optional
        The column name for the exposure times.  Values assumed to be in days.
        Default observationStartMJD.
    all_gaps : `bool`, optional
        Histogram the gaps between all observations (True) or just successive observations (False)?
        Default is False. If all gaps are used, this metric can become significantly slower.
    bins : `np.ndarray`, optional
        The bins to use for the histogram of time gaps (in days, or same units as times_col).
        Default values are bins from 0 to 2 hours, in 5 minute intervals.

    Returns
    -------
    histogram : `np.ndarray`
        Returns a histogram of the tgaps at each slice point;
        these histograms can be combined and plotted using the 'SummaryHistogram plotter'.
    """

    def __init__(
        self,
        times_col="observationStartMJD",
        all_gaps=False,
        bins=np.arange(0, 120.0, 5.0) / 60.0 / 24.0,
        units="days",
        **kwargs,
    ):
        # Pass the same bins to the plotter.
        self.bins = bins
        self.times_col = times_col
        super().__init__(col=[self.times_col], metric_dtype="object", units=units, **kwargs)
        self.all_gaps = all_gaps

    def run(self, data_slice, slice_point=None):
        if data_slice.size < 2:
            return self.badval
        times = np.sort(data_slice[self.times_col])
        if self.all_gaps:
            all_diffs = []
            for i in np.arange(1, times.size, 1):
                all_diffs.append((times - np.roll(times, i))[i:])
            dts = np.concatenate(all_diffs)
        else:
            dts = np.diff(times)
        result, bins = np.histogram(dts, self.bins)
        return result


class TgapsPercentMetric(BaseMetric):
    """Compute the fraction of the time gaps between observations that occur in a given time range.

    Measure the gaps between observations.  By default, only gaps
    between neighboring visits are computed.  If all_gaps is set to true, all gaps are
    computed (i.e., if there are observations at 10, 20, 30 and 40 the default will
    Compute the percent of gaps between specified endpoints.

    This is different from the TgapsMetric in that this only looks at what percent of intervals fall
    into the specified range, rather than histogramming the entire set of tgaps.

    Parameters
    ----------
    times_col : `str`, opt
        The column name for the exposure times.  Values assumed to be in days.
        Default observationStartMJD.
    all_gaps : `bool`, opt
        Histogram the gaps between all observations (True) or just successive observations (False)?
        Default is False. If all gaps are used, this metric can become significantly slower.
    min_time : `float`, opt
        Minimum time of gaps to include (days). Default 2/24 (2 hours).
    max_time : `float`, opt
        Max time of gaps to include (days). Default 14/24 (14 hours).

    Returns
    -------
    percent : `float`
        Returns a float percent of the CDF between cdfMinTime and cdfMaxTime -
        (# of tgaps within min_time/max_time / # of all tgaps).
    """

    def __init__(
        self,
        times_col="observationStartMJD",
        all_gaps=False,
        min_time=2.0 / 24,
        max_time=14.0 / 24,
        units="percent",
        **kwargs,
    ):
        self.times_col = times_col
        assert min_time <= max_time
        self.min_time = min_time
        self.max_time = max_time
        super().__init__(col=[self.times_col], metric_dtype="float", units=units, **kwargs)
        self.all_gaps = all_gaps

    def run(self, data_slice, slice_point=None):
        if data_slice.size < 2:
            return self.badval
        times = np.sort(data_slice[self.times_col])
        if self.all_gaps:
            all_diffs = []
            for i in np.arange(1, times.size, 1):
                all_diffs.append((times - np.roll(times, i))[i:])
            dts = np.concatenate(all_diffs)
        else:
            dts = np.diff(times)
        n_in_window = np.sum((dts >= self.min_time) & (dts <= self.max_time))
        return n_in_window / len(dts) * 100.0


class NightgapsMetric(BaseMetric):
    """Histogram the number of nights between observations.


    Measure the gaps between observations.  By default, only gaps
    between neighboring visits are computed.  If all_gaps is set to true, all gaps are
    computed (i.e., if there are observations at 10, 20, 30 and 40 the default will
    histogram [10,10,10] while all_gaps histograms [10,10,10,20,20,30])

    Parameters
    ----------
    night_col : `str`, optional
        The column name for the night of each observation.
        Default 'night'.
    all_gaps : `bool`, optional
        Histogram the gaps between all observations (True) or just successive observations (False)?
        Default is False. If all gaps are used, this metric can become significantly slower.
    bins : `np.ndarray`, optional
        The bins to use for the histogram of time gaps (in days, or same units as timesCol).
        Default values are bins from 0 to 10 days, in 1 day intervals.

    Returns
    -------
    histogram : `np.ndarray`
        Returns a histogram of the deltaT between nights at each slice point;
        these histograms can be combined and plotted using the 'SummaryHistogram plotter'.
    """

    def __init__(
        self,
        night_col="night",
        all_gaps=False,
        bins=np.arange(0, 10, 1),
        units="nights",
        **kwargs,
    ):
        # Pass the same bins to the plotter.
        self.bins = bins
        self.night_col = night_col
        super().__init__(col=[self.night_col], metric_dtype="object", units=units, **kwargs)
        self.all_gaps = all_gaps

    def run(self, data_slice, slice_point=None):
        if data_slice.size < 2:
            return self.badval
        nights = np.sort(np.unique(data_slice[self.night_col]))
        if self.all_gaps:
            all_diffs = []
            for i in np.arange(1, nights.size, 1):
                all_diffs.append((nights - np.roll(nights, i))[i:])
            dnights = np.concatenate(all_diffs)
        else:
            dnights = np.diff(nights)
        result, bins = np.histogram(dnights, self.bins)
        return result


class NVisitsPerNightMetric(BaseMetric):
    """Histogram the number of visits in each night.

    Splits the visits by night, then histograms how many visits occur in each night.

    Parameters
    ----------
    night_col : `str`, optional
        The column name for the night of each observation.
        Default 'night'.
    bins : `np.ndarray`, optional
        The bins to use for the histogram of time gaps (in days, or same units as timesCol).
        Default values are bins from 0 to 5 visits, in steps of 1.

    Returns
    -------
    histogram : `np.ndarray`
        Returns a histogram of the number of visits per night at each slice point;
        these histograms can be combined and plotted using the 'SummaryHistogram plotter'.
    """

    def __init__(self, night_col="night", bins=np.arange(0, 10, 1), units="#", **kwargs):
        # Pass the same bins to the plotter.
        self.bins = bins
        self.night_col = night_col
        super().__init__(col=[self.night_col], metric_dtype="object", units=units, **kwargs)

    def run(self, data_slice, slice_point=None):
        n, counts = np.unique(data_slice[self.night_col], return_counts=True)
        result, bins = np.histogram(counts, self.bins)
        return result


class MaxGapMetric(BaseMetric):
    """Find the maximum gap (in days) in between successive observations.

    Useful for making sure there is an image within the last year that would make a good template image.

    Parameters
    ----------
    mjd_col : `str`, opt
        The column name of the night of each observation.

    Returns
    -------
    maxGap : `float`
        The maximum gap (in days) between visits.
    """

    def __init__(self, mjd_col="observationStartMJD", **kwargs):
        self.mjd_col = mjd_col
        units = "Days"
        super(MaxGapMetric, self).__init__(col=[self.mjd_col], units=units, **kwargs)

    def run(self, data_slice, slice_point=None):
        gaps = np.diff(np.sort(data_slice[self.mjd_col]))
        if np.size(gaps) > 0:
            result = np.max(gaps)
        else:
            result = self.badval
        return result


class NightTimespanMetric(BaseMetric):
    """Calculate the maximum time span covered in each night, report the `percentile` value of all timespans.

    Parameters
    ----------
    percentile : `float`, opt
        Percentile value to report. Default 75th percentile.
    night_col : `str`, opt
        Name of the night column. Default 'night'.
    mjd_col : `str`, opt
        Name of the MJD visit column. Default 'observationStartMJD'.
    """

    def __init__(self, percentile=75, night_col="night", mjd_col="observationStartMJD", **kwargs):
        self.percentile = percentile
        self.night_col = night_col
        self.mjd_col = mjd_col
        if "metric_name" in kwargs:
            metric_name = kwargs["metric_name"]
            del kwargs["metric_name"]
        else:
            metric_name = f"{percentile}th Percentile Intranight Timespan"
        super().__init__(
            col=[self.night_col, self.mjd_col],
            units="minutes",
            metric_name=metric_name,
            **kwargs,
        )

    def run(self, data_slice, slice_point=None):
        data = np.sort(data_slice, order=self.mjd_col)
        unights, counts = np.unique(data[self.night_col], return_counts=True)
        unights = unights[np.where(counts > 1)]
        if len(unights) == 0:
            result = self.badval
        else:
            nstart = np.searchsorted(data[self.night_col], unights, side="left")
            nend = np.searchsorted(data[self.night_col], unights, side="right") - 1
            tspans = (data[self.mjd_col][nend] - data[self.mjd_col][nstart]) * 24.0 * 60.0
            result = np.percentile(tspans, self.percentile)
        return result



__all__ = ("obs2sqlite",)

import sqlite3
import sys

import healpy as hp
import numpy as np
import pandas as pd

import rubin_sim.skybrightness_pre as sb
from rubin_sim.skybrightness import SkyModel
from rubin_sim.utils import Site, _approx_ra_dec2_alt_az, m5_flat_sed, raDec2Hpid


def obs2sqlite(
    observations_in,
    location="LSST",
    outfile="observations.sqlite",
    slewtime_limit=5.0,
    full_sky=False,
    radians=True,
):
    """
    Utility to take an array of observations and dump it to a sqlite file, filling in useful columns along the way.

    observations_in: numpy array with at least columns of
        ra : RA in degrees
        dec : dec in degrees
        mjd : MJD in day
        filter : string with the filter name
        exptime : the exposure time in seconds
    slewtime_limit : float
        Consider all slewtimes larger than this to be closed-dome time not part of a slew.
    """

    # Set the location to be LSST
    if location == "LSST":
        telescope = Site("LSST")

    # Check that we have the columns we need
    needed_cols = ["ra", "dec", "mjd", "filter"]
    in_cols = observations_in.dtype.names
    for col in needed_cols:
        if needed_cols not in in_cols:
            ValueError("%s column not found in observtion array" % col)

    n_obs = observations_in.size
    sm = None
    # make sure they are in order by MJD
    observations_in.sort(order="mjd")

    # Take all the columns that are in the input and add any missing
    names = [
        "filter",
        "ra",
        "dec",
        "mjd",
        "exptime",
        "alt",
        "az",
        "skybrightness",
        "seeing",
        "night",
        "slewtime",
        "fivesigmadepth",
        "airmass",
        "sunAlt",
        "moonAlt",
    ]
    types = ["|S1"]
    types.extend([float] * (len(names) - 1))

    observations = np.zeros(n_obs, dtype=list(zip(names, types)))

    # copy over the ones we have
    for col in in_cols:
        observations[col] = observations_in[col]

    # convert output to be in degrees like expected
    if radians:
        observations["ra"] = np.degrees(observations["ra"])
        observations["dec"] = np.degrees(observations["dec"])

    if "exptime" not in in_cols:
        observations["exptime"] = 30.0

    # Fill in the slewtime. Note that filterchange time gets included in slewtimes
    if "slewtime" not in in_cols:
        # Assume MJD is midpoint of exposures
        mjd_sec = observations_in["mjd"] * 24.0 * 3600.0
        observations["slewtime"][1:] = (
            mjd_sec[1:]
            - mjd_sec[0:-1]
            - observations["exptime"][0:-1] * 0.5
            - observations["exptime"][1:] * 0.5
        )
        closed = np.where(observations["slewtime"] > slewtime_limit * 60.0)
        observations["slewtime"][closed] = 0.0

    # Let's just use the stupid-fast to get alt-az
    if "alt" not in in_cols:
        alt, az = _approx_ra_dec2_alt_az(
            np.radians(observations["ra"]),
            np.radians(observations["dec"]),
            telescope.latitude_rad,
            telescope.longitude_rad,
            observations["mjd"],
        )
        observations["alt"] = np.degrees(alt)
        observations["az"] = np.degrees(az)

    # Fill in the airmass
    if "airmass" not in in_cols:
        observations["airmass"] = 1.0 / np.cos(np.pi / 2.0 - np.radians(observations["alt"]))

    # Fill in the seeing
    if "seeing" not in in_cols:
        # XXX just fill in a dummy val
        observations["seeing"] = 0.8

    # Sky Brightness
    if "skybrightness" not in in_cols:
        if full_sky:
            sm = SkyModel(mags=True)
            for i, obs in enumerate(observations):
                sm.set_ra_dec_mjd(obs["ra"], obs["dec"], obs["mjd"], degrees=True)
                observations["skybrightness"][i] = sm.return_mags()[obs["filter"]]
        else:
            # Let's try using the pre-computed sky brighntesses
            sm = sb.SkyModelPre(preload=False)
            full = sm.return_mags(observations["mjd"][0])
            nside = hp.npix2nside(full["r"].size)
            imax = float(np.size(observations))
            for i, obs in enumerate(observations):
                indx = raDec2Hpid(nside, obs["ra"], obs["dec"])
                observations["skybrightness"][i] = sm.return_mags(obs["mjd"], indx=[indx])[obs["filter"]]
                sun_moon = sm.returnSunMoon(obs["mjd"])
                observations["sunAlt"][i] = sun_moon["sunAlt"]
                observations["moonAlt"][i] = sun_moon["moonAlt"]
                progress = i / imax * 100
                text = "\rprogress = %.2f%%" % progress
                sys.stdout.write(text)
                sys.stdout.flush()
            observations["sunAlt"] = np.degrees(observations["sunAlt"])
            observations["moonAlt"] = np.degrees(observations["moonAlt"])

    # 5-sigma depth
    for fn in np.unique(observations["filter"]):
        good = np.where(observations["filter"] == fn)
        observations["fivesigmadepth"][good] = m5_flat_sed(
            fn,
            observations["skybrightness"][good],
            observations["seeing"][good],
            observations["exptime"][good],
            observations["airmass"][good],
        )

    conn = sqlite3.connect(outfile)
    df = pd.DataFrame(observations)
    df.to_sql("observations", conn)



__all__ = ("WFDlabelStacker",)

import healpy as hp
import numpy as np

from .base_stacker import BaseStacker


class WFDlabelStacker(BaseStacker):
    """Add an 'areaId' column to flag whether a visit was inside the 'footprint'.

    Parameters
    ----------
    footprint : `np.NDarray`, optional
        The healpix map indicating the desired footprint region.
        If this is not defined (default None), then the entire sky is used as the footprint.
    fp_threshold : `float`, optional
        The threshold for the fraction of the visit area which falls within the footprint in order
        to be counted as 'in' the footprint. Default 0.4.
    areaIdName : `str`, optional
        Value to place in the areaId column for visits which match this area.
    ra_col : `str`, optional
        The name of the RA column. Default fieldRA.
    dec_col : `str`, optional
        The name of the Dec column. Default fieldDec.
    note_col : `str`, optional
        The name of the 'note' column in the database. Default 'note'. This is used to identify visits
        which were part of a DD sequence.
    exclude_dd : `bool`, optional
        Exclude (True) or include (False) visits which are part of a DD sequence within this 'area'.

    This stacker adds an areaId column in the opsim database, to be labelled with 'areaIdName' if the
    visit falls within the healpix footprint map and (optionally) is not tagged as a DD visit.
    If it falls outside the footprint, the visit is tagged as "NULL".
    If it was part of a DD sequence, the visit is tagged with an ID which is unique to that DD field,
    if 'exclude_dd' is True.
    Generally this would be likely to be used to tag visits as belonging to WFD - but not necessarily!
    Any healpix footprint is valid.
    """

    cols_added = ["area_id"]

    def __init__(
        self,
        footprint=None,
        fp_threshold=0.4,
        area_id_name="WFD",
        ra_col="fieldRA",
        dec_col="fieldDec",
        note_col="note",
        exclude_dd=True,
    ):
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.note_col = note_col
        self.cols_req = [self.ra_col, self.dec_col, self.note_col]
        self.cols_added_dtypes = [(str, 15)]
        self.units = [""]
        self.fp_threshold = fp_threshold
        self.area_id_name = area_id_name
        self.exclude_dd = exclude_dd
        if footprint is None:
            # If footprint was not defined, just set it to cover the entire sky, at nside=64
            footprint = np.ones(hp.nside2npix(64))
        self.footprint = footprint
        self.nside = hp.npix2nside(len(self.footprint))

    def define_ddname(self, note):
        field = note.replace("u,", "")
        field = field.split(",")[0].replace(",", "")
        return field

    def _run(self, sim_data, cols_present=False):
        # Even if cols_present is true, recalculate.
        # Set up DD names.
        d = set()
        for p in np.unique(sim_data[self.note_col]):
            if p.startswith("DD"):
                d.add(self.define_ddname(p))
        # Identify Healpixels associated with each visit.
        vec = hp.dir2vec(sim_data[self.ra_col], sim_data[self.dec_col], lonlat=True)
        vec = vec.swapaxes(0, 1)
        radius = np.radians(1.75)  # fov radius
        area_id = np.zeros(len(sim_data), self.cols_added_dtypes[0])
        for i, (v, note) in enumerate(zip(vec, sim_data[self.note_col])):
            # Identify the healpixels which would be inside this pointing
            pointing_healpix = hp.query_disc(self.nside, v, radius, inclusive=False)
            # The wfd_footprint consists of values of 0/1 if out/in WFD footprint
            hp_in_fp = self.footprint[pointing_healpix].sum()
            # So in_fp= the number of healpixels which were in the specified footprint
            # .. in the # in / total # > limit (0.4) then "yes" it's in the footprint
            in_fp = hp_in_fp / len(pointing_healpix)
            if note.startswith("DD") and self.exclude_dd:
                area_id[i] = self.define_ddname(note)
            else:
                if in_fp >= self.fp_threshold:
                    area_id[i] = self.area_id_name
                else:
                    area_id[i] = "NULL"
        sim_data["area_id"] = area_id
        return sim_data



"""A slicer that uses a Healpix grid to calculate metric values (at the center of each healpixel)."""

# Requires healpy
# See more documentation on healpy here http://healpy.readthedocs.org/en/latest/tutorial.html
# Also requires numpy (for histogram and power spectrum plotting)

__all__ = ("HealpixSlicer",)

import healpy as hp
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord

from rubin_sim.maf.plots.spatial_plotters import HealpixHistogram, HealpixSkyMap

from .base_spatial_slicer import BaseSpatialSlicer


class HealpixSlicer(BaseSpatialSlicer):
    """
    A spatial slicer that evaluates pointings on a healpix-based grid.

    Note that a healpix slicer is intended to evaluate the sky on a spatial grid.
    Usually this grid will be something like RA as the lon_col and Dec as the lat_col.
    However, it could also be altitude and azimuth, in which case altitude as lat_col,
    and azimuth as lon_col.
    An additional alternative is to use HA/Dec, with lon_col=HA, lat_col=Dec.

    When plotting with RA/Dec, the default HealpixSkyMap can be used, corresponding to
    {'rot': (0, 0, 0), 'flip': 'astro'}.
    When plotting with alt/az, either the LambertSkyMap can be used (if Basemap is installed)
    or the HealpixSkyMap can be used with a modified plot_dict,
    {'rot': (90, 90, 90), 'flip': 'geo'}.
    When plotting with HA/Dec, only the HealpixSkyMap can be used, with a modified plot_dict of
    {'rot': (0, -30, 0), 'flip': 'geo'}.

    Parameters
    ----------
    nside : `int`, optional
        The nside parameter of the healpix grid. Must be a power of 2.
        Default 128.
    lon_col : `str`, optional
        Name of the longitude (RA equivalent) column to use from the input data.
        Default fieldRA
    lat_col : `str`, optional
        Name of the latitude (Dec equivalent) column to use from the input data.
        Default fieldDec
    lat_lon_deg : `bool`, optional
        Flag indicating whether the lat and lon values in the input data are in
        degrees (True) or radians (False).
        Default True.
    verbose : `bool`, optional
        Flag to indicate whether or not to write additional information to stdout during runtime.
        Default True.
    badval : `float`, optional
        Bad value flag, relevant for plotting. Default the np.nan value (in order to properly flag
        bad data points for plotting with the healpix plotting routines). This should not be changed.
    use_cache : `bool`, optional
        Flag allowing the user to indicate whether or not to cache (and reuse) metric results
        calculated with the same set of simulated data pointings.
        This can be safely set to True for slicers not using maps and will result in increased speed.
        When calculating metric results using maps, the metadata at each individual ra/dec point may
        influence the metric results and so use_cache should be set to False.
        Default True.
    leafsize : `int`, optional
        Leafsize value for kdtree. Default 100.
    radius : `float`, optional
        Radius for matching in the kdtree. Equivalent to the radius of the FOV. Degrees.
        Default 2.45.
    use_camera : `bool`, optional
        Flag to indicate whether to use the LSST camera footprint or not.
        Default True.
    camera_footprint_file : `str`, optional
        Name of the camera footprint map to use. Can be None, which will use the default.
    rot_sky_pos_col_name : `str`, optional
        Name of the rotSkyPos column in the input  data. Only used if use_camera is True.
        Describes the orientation of the camera orientation compared to the sky.
        Default rotSkyPos.
    """

    def __init__(
        self,
        nside=128,
        lon_col="fieldRA",
        lat_col="fieldDec",
        lat_lon_deg=True,
        verbose=True,
        badval=np.nan,
        use_cache=True,
        leafsize=100,
        radius=2.45,
        use_camera=True,
        camera_footprint_file=None,
        rot_sky_pos_col_name="rotSkyPos",
    ):
        """Instantiate and set up healpix slicer object."""
        super().__init__(
            verbose=verbose,
            lon_col=lon_col,
            lat_col=lat_col,
            badval=badval,
            radius=radius,
            leafsize=leafsize,
            use_camera=use_camera,
            camera_footprint_file=camera_footprint_file,
            rot_sky_pos_col_name=rot_sky_pos_col_name,
            lat_lon_deg=lat_lon_deg,
        )
        # Valid values of nside are powers of 2.
        # nside=64 gives about 1 deg resolution
        # nside=256 gives about 13' resolution (~1 CCD)
        # nside=1024 gives about 3' resolution
        # Check validity of nside:
        if not (hp.isnsideok(nside)):
            raise ValueError("Valid values of nside are powers of 2.")
        self.nside = int(nside)
        self.pix_area = hp.nside2pixarea(self.nside)
        self.nslice = hp.nside2npix(self.nside)
        self.shape = self.nslice
        if self.verbose:
            print(
                "Healpix slicer using NSIDE=%d, " % (self.nside)
                + "approximate resolution %f arcminutes" % (hp.nside2resol(self.nside, arcmin=True))
            )
        # Set variables so slicer can be re-constructed
        self.slicer_init = {
            "nside": nside,
            "lon_col": lon_col,
            "lat_col": lat_col,
            "radius": radius,
        }
        self.use_cache = use_cache
        if use_cache:
            # use_cache set the size of the cache for the memoize function in sliceMetric.
            bin_res = hp.nside2resol(nside)  # Pixel size in radians
            # Set the cache size to be ~2x the circumference
            self.cache_size = int(np.round(4.0 * np.pi / bin_res))
        # Set up slice_point metadata.
        self.slice_points["nside"] = nside
        self.slice_points["sid"] = np.arange(self.nslice)
        self.slice_points["ra"], self.slice_points["dec"] = self._pix2radec(self.slice_points["sid"])
        c = SkyCoord(ra=self.slice_points["ra"] * u.rad, dec=self.slice_points["dec"] * u.rad).transform_to(
            "galactic"
        )
        gall, galb = c.l.rad, c.b.rad

        self.slice_points["gall"] = gall
        self.slice_points["galb"] = galb

        # Set the default plotting functions.
        self.plot_funcs = [HealpixSkyMap, HealpixHistogram]

    def __eq__(self, other_slicer):
        """Evaluate if two slicers are equivalent."""
        # If the two slicers are both healpix slicers, check nsides value.
        result = False
        if isinstance(other_slicer, HealpixSlicer):
            if other_slicer.nside == self.nside:
                if other_slicer.lon_col == self.lon_col and other_slicer.lat_col == self.lat_col:
                    if other_slicer.radius == self.radius:
                        if other_slicer.use_camera == self.use_camera:
                            if other_slicer.rotSkyPosColName == self.rotSkyPosColName:
                                if np.all(other_slicer.shape == self.shape):
                                    if other_slicer.use_cache == self.use_cache:
                                        result = True
        return result

    def _pix2radec(self, islice):
        """Given the pixel number / sliceID, return the RA/Dec of the pointing, in radians."""
        # Calculate RA/Dec in RADIANS of pixel in this healpix slicer.
        # Note that ipix could be an array,
        # in which case RA/Dec values will be an array also.
        lat, ra = hp.pix2ang(self.nside, islice)
        # Move dec to +/- 90 degrees
        dec = np.pi / 2.0 - lat
        return ra, dec



# Example of a new metric added to the repo.
# ljones@astro.washington.edu

__all__ = ("NightsWithNFiltersMetric",)

import numpy as np

from rubin_sim.maf.metrics import BaseMetric


class NightsWithNFiltersMetric(BaseMetric):
    """Count how many times more than NFilters are used within the same night, for this set of visits."""

    def __init__(self, night_col="night", filter_col="filter", n_filters=3, **kwargs):
        """
        night_col = the name of the column defining the night
        filter_col = the name of the column defining the filter
        n_filters = the minimum desired set of filters used in these visits
        """
        self.night_col = night_col
        self.filter_col = filter_col
        self.n_filters = n_filters
        super(NightsWithNFiltersMetric, self).__init__(col=[self.night_col, self.filter_col], **kwargs)

    def run(self, data_slice, slice_point=None):
        count = 0
        unique_nights = np.unique(data_slice[self.night_col])
        for n in unique_nights:
            condition = data_slice[self.night_col] == n
            unique_filters = np.unique(data_slice[self.filter_col][condition])
            if len(unique_filters) > self.n_filters:
                count += 1
        return count



# Example for PeriodicMetric
# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 3/10/2015
# Motivation: The detection of periodic signals can be examined by using canonical signals and attempted
# to recover these. However, a more general approach would be to examine the strength in signal that is
# lost as a result of poor phase coverage.
# This metric calculates the spectral window function for a set of scheduled observations.
# The largest peak at a nonzero frequency is used as a proxy to quantify how much power is
# lost to other frequencies. Full phase coverage will result in a value of 1.
# We refer to this as the Periodic Purity Function.

__all__ = ("PeriodicMetric",)

import numpy as np

from rubin_sim.maf.metrics import BaseMetric


class PeriodicMetric(BaseMetric):
    """From a set of observation times, uses code provided by Robert Siverd (LCOGT) to calculate
    the spectral window function.
    """

    def __init__(self, time_col="expMJD", **kwargs):
        self.time_col = time_col
        super(PeriodicMetric, self).__init__(col=[self.time_col], **kwargs)

    def run(self, data_slice, slice_point=None):
        frq_pts = 30000.0
        max_frq = 25.0
        times = data_slice[self.time_col]
        times = times - times[0]  # change times to smaller numbers
        use_jd = np.array(times)
        window_frq = np.arange(frq_pts) * max_frq / frq_pts
        window_val = np.zeros_like(window_frq, dtype="float")
        for x, frq in enumerate(window_frq):
            window_val[x] = np.sum(np.cos(-2.0 * np.pi * frq * use_jd))
        window_val /= np.float(use_jd.size)
        secondpeak = np.sort(window_val)[-2]
        totalsum = (np.sum(window_val) - np.sort(window_val)[-1]) / (frq_pts - 1)
        data = np.asarray([secondpeak, totalsum])
        return data

    def reduce_peak(self, data):
        return 1.0 - data[0]

    def reduce_sum(self, data):
        return data[1]



"""Some basic physical quantity metrics.
"""
__all__ = (
    "metadataBasics",
    "metadataBasicsAngle",
    "allMetadata",
    "metadataMaps",
    "firstYearMetadata",
)

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers
import rubin_sim.maf.stackers as stackers

from .col_map_dict import col_map_dict
from .common import (
    combine_info_labels,
    extended_metrics,
    filter_list,
    standard_angle_metrics,
    standard_summary,
)


def metadataBasics(
    value,
    colmap=None,
    runName="opsim",
    valueName=None,
    groupName=None,
    extraSql=None,
    extraInfoLabel=None,
    slicer=None,
):
    """Calculate basic metrics on visit metadata 'value'
    (e.g. airmass, normalized airmass, seeing..).
    Calculates this around the sky (HealpixSlicer),
    makes histograms of all visits (OneDSlicer),
    and calculates statistics on all visits (UniSlicer) for the quantity,
     in all visits and per filter.

    Currently have a hack for HA & normairmass.

    Parameters
    ----------
    value : `str`
        The column name for the quantity to evaluate.
         (column name in the database or created by a stacker).
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    valueName : `str`, optional
        The name of the value to be reported in the results_db and
        added to the metric.
        This is intended to help standardize metric comparison between
        sim versions.
        value = name as it is in the database (seeingFwhmGeom, etc).
        valueName = name to be recorded ('seeingGeom', etc.).
    groupName : `str`, optional
        The group name for this quantity in the display_dict.
        None will default to the same as 'valueName', capitalized.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints.
    extraInfoLabel : `str`, optional
        Additional info_labels to add before any below (i.e. "WFD").
    slicer : `rubin_sim.maf.slicers.BaseSlicer`, optional
        Optionally use a different slicer than an nside=64 healpix slicer.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    if valueName is None:
        valueName = value

    if groupName is None:
        groupName = valueName.capitalize()
        subgroup = extraInfoLabel
    else:
        groupName = groupName.capitalize()
        subgroup = valueName.capitalize()

    if subgroup is None:
        subgroup = "All visits"

    displayDict = {"group": groupName, "subgroup": subgroup}

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]

    # Set up basic all and per filter sql constraints.
    filterlist, colors, orders, sqls, info_label = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=extraInfoLabel
    )

    if slicer is not None:
        skyslicer = slicer
    else:
        skyslicer = slicers.HealpixSlicer(nside=64, lon_col=raCol, lat_col=decCol, lat_lon_deg=degrees)

    # Hack to make HA work, but really need to account for any stackers
    if value == "HA":
        stackerList = [stackers.HourAngleStacker(lst_col=colmap["lst"], ra_col=raCol, degrees=degrees)]
    elif value == "normairmass":
        stackerList = [stackers.NormAirmassStacker(degrees=degrees)]
    else:
        stackerList = None

    # Summarize values over all and per filter
    # (min/mean/median/max/percentiles/outliers/rms).
    slicer = slicers.UniSlicer()
    for f in filterlist:
        for m in extended_metrics(value, replace_colname=valueName):
            displayDict["caption"] = "%s for %s." % (m.name, info_label[f])
            displayDict["order"] = orders[f]
            bundle = mb.MetricBundle(
                m,
                slicer,
                sqls[f],
                stacker_list=stackerList,
                info_label=info_label[f],
                display_dict=displayDict,
            )
            bundleList.append(bundle)

    # Histogram values over all and per filter.
    for f in filterlist:
        displayDict["caption"] = "Histogram of %s" % (value)
        if valueName != value:
            displayDict["caption"] += " (%s)" % (valueName)
        displayDict["caption"] += " for %s." % (info_label[f])
        displayDict["order"] = orders[f]
        m = metrics.CountMetric(value, metric_name="%s Histogram" % (valueName))
        slicer = slicers.OneDSlicer(slice_col_name=value)
        bundle = mb.MetricBundle(
            m,
            slicer,
            sqls[f],
            stacker_list=stackerList,
            info_label=info_label[f],
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    # Make maps of min/median/max for all and per filter,
    # per RA/Dec, with standard summary stats.
    plotDict = {"percentile_clip": 98}
    mList = []
    mList.append(metrics.MinMetric(value, metric_name="Min %s" % (valueName)))
    mList.append(metrics.MedianMetric(value, metric_name="Median %s" % (valueName)))
    mList.append(metrics.MaxMetric(value, metric_name="Max %s" % (valueName)))
    slicer = skyslicer
    for f in filterlist:
        for m in mList:
            displayDict["caption"] = "Map of %s" % m.name
            if valueName != value:
                displayDict["caption"] += " (%s)" % value
            displayDict["caption"] += " for %s." % info_label[f]
            displayDict["order"] = orders[f]
            bundle = mb.MetricBundle(
                m,
                slicer,
                sqls[f],
                stacker_list=stackerList,
                info_label=info_label[f],
                plot_dict=plotDict,
                display_dict=displayDict,
                summary_metrics=standard_summary(),
            )
            bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)


def metadataBasicsAngle(
    value,
    colmap=None,
    runName="opsim",
    valueName=None,
    groupName=None,
    extraSql=None,
    extraInfoLabel=None,
    slicer=None,
):
    """Calculate basic metrics on visit metadata 'value',
    where value is a wrap-around angle.

    Calculates extended standard metrics (with unislicer) on the quantity
    (all visits and per filter),
    makes histogram of the value (all visits and per filter),


    Parameters
    ----------
    value : `str`
        The column name for the quantity to evaluate.
        (column name in the database or created by a stacker).
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    valueName : `str`, optional
        The name of the value to be reported in the results_db
        and added to the metric.
        This is intended to help standardize metric comparison
        between sim versions.
        value = name as it is in the database (seeingFwhmGeom, etc).
        valueName = name to be recorded ('seeingGeom', etc.).
    groupName : `str`, optional
        The group name for this quantity in the display_dict.
        None will default to the same as 'valueName', capitalized.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints.
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").
    slicer : `rubin_sim.maf.slicer.BaseSlicer` or None, optional

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    if valueName is None:
        valueName = value

    if groupName is None:
        groupName = valueName.capitalize()
        subgroup = extraInfoLabel
    else:
        groupName = groupName.capitalize()
        subgroup = valueName.capitalize()

    if subgroup is None:
        subgroup = "All visits"

    displayDict = {"group": groupName, "subgroup": subgroup}

    raCol = colmap["ra"]
    decCol = colmap["dec"]
    degrees = colmap["raDecDeg"]
    # Set up basic all and per filter sql constraints.
    filterlist, colors, orders, sqls, info_label = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=extraInfoLabel
    )

    if slicer is not None:
        skyslicer = slicer
    else:
        skyslicer = slicers.HealpixSlicer(nside=64, lon_col=raCol, lat_col=decCol, lat_lon_deg=degrees)

    # Summarize values over all and per filter.
    slicer = slicers.UniSlicer()
    for f in filterlist:
        for m in standard_angle_metrics(value, replace_colname=valueName):
            displayDict["caption"] = "%s for %s." % (m.name, info_label[f])
            displayDict["order"] = orders[f]
            bundle = mb.MetricBundle(
                m,
                slicer,
                sqls[f],
                info_label=info_label[f],
                display_dict=displayDict,
            )
            bundleList.append(bundle)

    # Histogram values over all and per filter.
    for f in filterlist:
        displayDict["caption"] = "Histogram of %s" % (value)
        if valueName != value:
            displayDict["caption"] += " (%s)" % (valueName)
        displayDict["caption"] += " for %s." % (info_label[f])
        displayDict["order"] = orders[f]
        m = metrics.CountMetric(value, metric_name="%s Histogram" % (valueName))
        slicer = slicers.OneDSlicer(slice_col_name=value)
        bundle = mb.MetricBundle(
            m,
            slicer,
            sqls[f],
            info_label=info_label[f],
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    # Make maps of min/median/max for all and per filter,
    # per RA/Dec, with standard summary stats.
    mList = []
    mList.append(metrics.MeanAngleMetric(value, metric_name="AngleMean %s" % (valueName)))
    mList.append(metrics.FullRangeAngleMetric(value, metric_name="AngleRange %s" % (valueName)))
    mList.append(metrics.RmsAngleMetric(value, metric_name="AngleRms %s" % (valueName)))
    plotDict = {"percentile_clip": 98}
    slicer = skyslicer
    for f in filterlist:
        for m in mList:
            displayDict["caption"] = "Map of %s" % m.name
            if valueName != value:
                displayDict["caption"] += " (%s)" % value
            displayDict["caption"] += " for %s." % info_label[f]
            displayDict["order"] = orders[f]
            bundle = mb.MetricBundle(
                m,
                slicer,
                sqls[f],
                info_label=info_label[f],
                plot_dict=plotDict,
                display_dict=displayDict,
                summary_metrics=standard_summary(),
            )
            bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)


def allMetadata(colmap=None, runName="opsim", extraSql=None, extraInfoLabel=None, slicer=None):
    """Generate a large set of metrics about the metadata of each visit -
    distributions of airmass, normalized airmass, seeing, sky brightness,
    single visit depth, hour angle, distance to the moon, and solar elongation.
    The exact metadata which is analyzed is set by the colmap['metadataList'].

    Parameters
    ----------
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    extraSql : `str`, optional
        Sql constraint (such as WFD only).
    extraInfoLabel : `str`, optional
        Metadata to identify the sql constraint (such as WFD).
    slicer : `rubin_sim.maf.slicer.BaseSlicer` or None, optional
        Optionally use something other than an nside=64 healpix slicer.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """

    if colmap is None:
        colmap = col_map_dict()

    bdict = {}

    for valueName in colmap["metadataList"]:
        if valueName in colmap:
            value = colmap[valueName]
        else:
            value = valueName
        mdict = metadataBasics(
            value,
            colmap=colmap,
            runName=runName,
            valueName=valueName,
            extraSql=extraSql,
            extraInfoLabel=extraInfoLabel,
            slicer=slicer,
        )
        bdict.update(mdict)
    for valueName in colmap["metadataAngleList"]:
        if valueName in colmap:
            value = colmap[valueName]
        else:
            value = valueName
        mdict = metadataBasicsAngle(
            value,
            colmap=colmap,
            runName=runName,
            valueName=valueName,
            extraSql=extraSql,
            extraInfoLabel=extraInfoLabel,
            slicer=slicer,
        )
        bdict.update(mdict)
    return bdict


def metadataMaps(
    value,
    colmap=None,
    runName="opsim",
    valueName=None,
    groupName=None,
    extraSql=None,
    extraInfoLabel=None,
    slicer=None,
):
    """Calculate 25/50/75 percentile values on maps across sky
    for a single metadata value.

    Parameters
    ----------
    value : `str`
        The column name for the quantity to evaluate.
        (column name in the database or created by a stacker).
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    valueName : `str`, optional
        The name of the value to be reported in the results_db
        and added to the metric.
        This is intended to help standardize metric comparison
        between sim versions.
        value = name as it is in the database (seeingFwhmGeom, etc).
        valueName = name to be recorded ('seeingGeom', etc.).
    groupName : `str`, optional
        The group name for this quantity in the display_dict.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints.
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").
    slicer : `rubin_sim.maf.slicer.BaseSlicer` or None, optional
        Optionally use something other than an nside=64 HealpixSlicer

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    if valueName is None:
        valueName = value

    if groupName is None:
        groupName = valueName.capitalize()
        subgroup = extraInfoLabel
    else:
        groupName = groupName.capitalize()
        subgroup = valueName.capitalize()

    if subgroup is None:
        subgroup = "All visits"

    displayDict = {"group": groupName, "subgroup": subgroup}

    ra_col = colmap["ra"]
    dec_col = colmap["dec"]
    degrees = colmap["raDecDeg"]
    # Set up basic all and per filter sql constraints.
    filterlist, colors, orders, sqls, info_label = filter_list(
        all=True, extra_sql=extraSql, extra_info_label=extraInfoLabel
    )

    # Hack to make HA work, but really need to account for any stackers
    if value == "HA":
        stackerList = [stackers.HourAngleStacker(lst_col=colmap["lst"], ra_col=ra_col, degrees=degrees)]
    elif value == "normairmass":
        stackerList = [stackers.NormAirmassStacker(degrees=degrees)]
    else:
        stackerList = None

    # Make maps of 25/median/75 for all and per filter,
    # per RA/Dec, with standard summary stats.
    mList = []
    mList.append(
        metrics.PercentileMetric(value, percentile=25, metric_name="25thPercentile %s" % (valueName))
    )
    mList.append(metrics.MedianMetric(value, metric_name="Median %s" % (valueName)))
    mList.append(
        metrics.PercentileMetric(value, percentile=75, metric_name="75thPercentile %s" % (valueName))
    )
    if slicer is None:
        slicer = slicers.HealpixSlicer(nside=64, lat_col=dec_col, lon_col=ra_col, lat_lon_deg=degrees)

    for f in filterlist:
        for m in mList:
            displayDict["caption"] = "Map of %s" % m.name
            if valueName != value:
                displayDict["caption"] += " (%s)" % value
            displayDict["caption"] += " for %s." % info_label[f]
            displayDict["order"] = orders[f]
            bundle = mb.MetricBundle(
                m,
                slicer,
                sqls[f],
                stacker_list=stackerList,
                info_label=info_label[f],
                display_dict=displayDict,
                summary_metrics=standard_summary(),
            )
            bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)

    return mb.make_bundles_dict_from_list(bundleList)


def firstYearMetadata(colmap=None, runName="opsim", extraSql=None, extraInfoLabel=None, slicer=None):
    """Measure the distribution of some basic metadata in the first year
    of operations -
    distributions of airmass, seeing, sky brightness, single visit depth.

    Parameters
    ----------
    colmap : `dict` or None, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    extraSql : `str`, optional
        Sql constraint (such as WFD only).
    extraInfoLabel : `str`, optional
        Metadata to identify the sql constraint (such as WFD).
    slicer : `rubin_sim.maf.slicer.BaseSlicer` or None, optional
        Optionally use something other than an nside=64 healpix slicer.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """

    if colmap is None:
        colmap = col_map_dict()

    bdict = {}

    firstYr = "night < 365.5"
    if extraSql is not None:
        extraSql = f"({firstYr}) and ({extraSql})"
    else:
        extraSql = firstYr
    extraInfoLabel = combine_info_labels("Yr 1", extraInfoLabel)

    subset = ["airmass", "seeingEff", "seeingGeom", "skyBrightness", "fiveSigmaDepth"]
    for valueName in subset:
        if valueName in colmap:
            value = colmap[valueName]
        else:
            value = valueName
        mdict = metadataBasics(
            value,
            colmap=colmap,
            runName=runName,
            valueName=valueName,
            extraSql=extraSql,
            extraInfoLabel=extraInfoLabel,
            slicer=slicer,
        )
        bdict.update(mdict)

    return bdict



import unittest

import numpy as np
from rubin_scheduler.utils import survey_start_mjd

from rubin_sim.satellite_constellations import Constellation, oneweb_tles, starlink_tles_v1, starlink_tles_v2


class TestSatellites(unittest.TestCase):
    def test_constellations(self):
        """Test stellite constellations"""

        mjd0 = survey_start_mjd()
        sv1 = starlink_tles_v1()
        sv2 = starlink_tles_v2()
        ow = oneweb_tles()

        const = Constellation(sv1)

        lengths, n_s = const.check_pointings(
            np.array([85.0, 82.0]),
            np.array([0.0, 0.0]),
            np.arange(2) + mjd0 + 1.5,
            30.0,
        )

        assert np.size(lengths) == 2
        assert np.size(n_s) == 2


if __name__ == "__main__":
    unittest.main()



__all__ = (
    "power_law_dndh",
    "neo_dndh_granvik",
    "neo_dndh_grav",
    "pha_dndh_granvik",
    "pha_dndh_grav",
    "integrate_over_h",
    "sum_over_h",
    "TotalNumberSSO",
    "ValueAtHMetric",
    "MeanValueAtHMetric",
    "MoCompletenessMetric",
    "MoCompletenessAtTimeMetric",
)

import warnings

import numpy as np

from .mo_metrics import BaseMoMetric


def power_law_dndh(hvalues, hindex=0.33, no=None, ho=None, **kwargs):
    """Power law distribution of objects.

    Parameters
    ----------
    hvalues : `np.ndarray`, (N,)
        The H values corresponding to each metric_value
        (must be the same length).
        The hvalues are expected to be evenly spaced.
    hindex : `float`, optional
        The power-law index expected for the H value distribution.
        Default is 0.33  (dN/dH = 10^(hindex * H) ).
    no : `float`, optional
    ho: `float`, optional
        If no and ho are specified, this provides an anchor
        for the power law distribution,so that the expected number no
        of objects at ho is returned.
        Does not need to be set if just doing comparative weighting.

    Returns
    -------
    dndh : `np.ndarray`, (N,)
    """
    if no is None or ho is None:
        ho = hvalues.min()
        no = 10
    binratio = (np.diff(hvalues, append=hvalues[-1] + np.diff(hvalues)[-1])) / 0.1
    dndh = (no * 0.1) * np.power(10.0, hindex * (hvalues - ho)) * binratio
    return dndh


def neo_dndh_granvik(hvalues, **kwargs):
    binratio = (np.diff(hvalues, append=hvalues[-1] + np.diff(hvalues)[-1])) / 0.1
    y0 = 0  # 10 * np.power(10, 0.55 * (x - 17))
    y1 = 150 * np.power(10, 0.3 * (hvalues - 18.5))
    y2 = 2500 * np.power(10, 0.92 * (hvalues - 23.2))
    dndh = (y0 + y1 + y2) * binratio
    return dndh


def neo_dndh_grav(hvalues, **kwargs):
    binratio = (np.diff(hvalues, append=hvalues[-1] + np.diff(hvalues)[-1])) / 0.1
    y1 = 110 * np.power(10, 0.35 * (hvalues - 18.5))
    dndh = y1 * binratio
    return dndh


def pha_dndh_granvik(hvalues, **kwargs):
    binratio = (np.diff(hvalues, append=hvalues[-1] + np.diff(hvalues)[-1])) / 0.1
    y0 = 0  # 10 * np.power(10, 0.55 * (x - 17))
    y1 = 20 * np.power(10, 0.3 * (hvalues - 18.5))
    y2 = 500 * np.power(10, 0.92 * (hvalues - 23.2))
    dndh = (y0 + y1 + y2) * binratio
    return dndh


def pha_dndh_grav(hvalues, **kwargs):
    binratio = (np.diff(hvalues, append=hvalues[-1] + np.diff[hvalues][-1])) / 0.1
    y1 = 23.5 * np.power(10, 0.35 * (hvalues - 18.5))
    dndh = y1 * binratio
    return dndh


def integrate_over_h(metric_values, hvalues, dndh_func=power_law_dndh, **kwargs):
    """Calculate a metric value integrated over an h_range.
    This is the metric value *weighted* by the size distribution.

    Parameters
    ----------
    metric_values : `numpy.ndarray`
        The metric values at each H value.
    hvalues : `numpy.ndarray`
        The H values corresponding to each metric_value
        (must be the same length).
    dndh_func : function, optional
        One of the dN/dH functions defined below.
    **kwargs : `dict`, optional
        Keyword arguments to pass to dndh_func

    Returns
    --------
    int_vals : `np.ndarray`, (N,)
       The integrated metric values.
    """
    # Set expected H distribution.
    # dndh = differential size distribution (number in this bin)
    dndh = dndh_func(hvalues, **kwargs)
    # calculate the metric values *weighted* by the number of objects
    # in this bin and brighter
    int_vals = np.cumsum(metric_values * dndh) / np.cumsum(dndh)
    return int_vals


def sum_over_h(metric_values, hvalues, dndh_func=power_law_dndh, **kwargs):
    """Calculate the sum of the metric value multiplied by the number of
    objects at each H value. This is equivalent to calculating the
    number of objects meeting X requirement in the differential completeness
    or fraction of objects with lightcurves, etc.

    Parameters
    ----------
    metric_values : `np.ndarray`, (N,)
        The metric values at each H value.
    hvalues : `np.ndarray`, (N,)
        The H values corresponding to each metric_value.
    dndh_func : function, optional
        One of the dN/dH functions defined below.
    **kwargs : `dict`, optional
        Keyword arguments to pass to dndh_func

    Returns
    --------
    sum_vals : `np.ndarray`, (N,)
       The cumulative metric values.
    """
    # Set expected H distribution.
    # dndh = differential size distribution (number in this bin)
    dndh = dndh_func(hvalues, **kwargs)
    # calculate the metric values *weighted* by the number of objects
    # in this bin and brighter
    sum_vals = np.cumsum(metric_values * dndh)
    return sum_vals


class TotalNumberSSO(BaseMoMetric):
    """Calculate the total number of objects of a given population
    expected at a given H value or larger.

    Operations on differential completeness values
    (or equivalent; fractions of the population is ok if
    still a differential metric result, not cumulative).

    Parameters
    ----------
    h_mark : `float`, optional
        The H value at which to calculate the expected total number of objects.
    dndh_func : function, optional
        The dN/dH distribution used calculate the expected population size.

    Returns
    -------
    nObj : `float`
        The predicted number of objects in the population.
    """

    def __init__(self, h_mark=22, dndh_func=neo_dndh_granvik, **kwargs):
        self.h_mark = h_mark
        self.dndh_func = dndh_func
        metric_name = "Nobj <= %.1f" % (h_mark)
        self.kwargs = kwargs
        super().__init__(metric_name=metric_name, **kwargs)

    def run(self, metric_vals, h_vals):
        totals = sum_over_h(metric_vals, h_vals, self.dndh_func, **self.kwargs)
        n_obj = np.interp(self.h_mark, h_vals, totals)
        return n_obj


class ValueAtHMetric(BaseMoMetric):
    """Return the metric value at a given H value.

    Requires the metric values to be one-dimensional
    (typically, completeness values).

    Parameters
    ----------
    h_mark : `float`, optional
        The H value at which to look up the metric value.

    Returns
    -------
    value: : `float`
    """

    def __init__(self, h_mark=22, **kwargs):
        metric_name = "Value At H=%.1f" % (h_mark)
        self.units = "<= %.1f" % (h_mark)
        super().__init__(metric_name=metric_name, **kwargs)
        self.h_mark = h_mark

    def run(self, metric_vals, h_vals):
        # Check if desired H value is within range of H values.
        if (self.h_mark < h_vals.min()) or (self.h_mark > h_vals.max()):
            warnings.warn("Desired H value of metric outside range of provided H values.")
            return None
        if metric_vals.shape[0] != 1:
            warnings.warn("This is not an appropriate summary statistic for this data - need 1d values.")
            return None
        value = np.interp(self.h_mark, h_vals, metric_vals[0])
        return value


class MeanValueAtHMetric(BaseMoMetric):
    """Return the mean value of a metric at a given H.

    Allows the metric values to be multi-dimensional
    (i.e. use a cloned H distribution).

    Parameters
    ----------
    h_mark : `float`, optional
        The H value at which to look up the metric value.

    Returns
    -------
    value: : `float`
    """

    def __init__(self, h_mark=22, reduce_func=np.mean, metric_name=None, **kwargs):
        if metric_name is None:
            metric_name = "Mean Value At H=%.1f" % (h_mark)
        self.units = "@ H= %.1f" % (h_mark)
        super().__init__(metric_name=metric_name, **kwargs)
        self.h_mark = h_mark
        self.reduce_func = reduce_func

    def run(self, metric_vals, h_vals):
        # Check if desired H value is within range of H values.
        if (self.h_mark < h_vals.min()) or (self.h_mark > h_vals.max()):
            warnings.warn("Desired H value of metric outside range of provided H values.")
            return None
        value = np.interp(self.h_mark, h_vals, self.reduce_func(metric_vals.swapaxes(0, 1), axis=1))
        return value


class MoCompletenessMetric(BaseMoMetric):
    """Calculate the fraction of the population that meets `threshold` value
    or higher. This is equivalent to calculating the completeness
    (relative to the entire population) given the output of a
    Discovery_N_Chances metric, or the fraction of the population that meets
    a given cutoff value for Color determination metrics.

    Any moving object metric that outputs a float value can thus have
    the 'fraction of the population' with greater than X value calculated here,
    as a summary statistic.

    Parameters
    ----------
    threshold : `int`, optional
        Count the fraction of the population that exceeds this value.
    nbins : `int`, optional
        If the H values for the metric are not a cloned distribution,
        then split up H into this many bins.
    min_hrange : `float`, optional
        If the H values for the metric are not a cloned distribution,
        then split up H into at least this
        range (otherwise just use the min/max of the H values).
    cumulative : `bool`, optional
        If False, simply report the differential fractional value
        (or differential completeness).
        If True, integrate over the H distribution (using IntegrateOverH)
        to report a cumulative fraction.
        Default of None will use True, unless metric_name is set and starts
        with "Differential" - then default will use False.
    hindex : `float`, optional
        Use hindex as the power law to integrate over H,
        if cumulative is True.
    """

    def __init__(
        self,
        threshold=1,
        nbins=20,
        min_hrange=1.0,
        cumulative=None,
        hindex=0.33,
        **kwargs,
    ):
        if cumulative is None:
            # if metric_name does not start with 'differential',
            # then cumulative->True
            if "metric_name" not in kwargs:
                self.cumulative = True
                metric_name = "CumulativeCompleteness"
            else:
                #  'metric_name' in kwargs:
                metric_name = kwargs.pop("metric_name")
                if metric_name.lower().startswith("differential"):
                    self.cumulative = False
                else:
                    self.cumulative = True
        else:
            # cumulative was set
            self.cumulative = cumulative
            if "metric_name" in kwargs:
                metric_name = kwargs.pop("metric_name")
                if metric_name.lower().startswith("differential") and self.cumulative:
                    warnings.warn(f"Completeness metric_name is {metric_name} but cumulative is True")
            else:
                if self.cumulative:
                    metric_name = "CumulativeCompleteness"
                else:
                    metric_name = "DifferentialCompleteness"
        if self.cumulative:
            units = "<=H"
        else:
            units = "@H"
        super().__init__(metric_name=metric_name, units=units, **kwargs)
        self.threshold = threshold
        # If H is not a cloned distribution,
        # then we need to specify how to bin these values.
        self.nbins = nbins
        self.min_hrange = min_hrange
        self.hindex = hindex

    def run(self, metric_values, h_vals):
        n_ssos = metric_values.shape[0]
        n_hval = len(h_vals)
        metric_val_h = metric_values.swapaxes(0, 1)
        if n_hval == metric_values.shape[1]:
            # h_vals array is probably the same as the cloned H array.
            completeness = np.zeros(len(h_vals), float)
            for i, H in enumerate(h_vals):
                completeness[i] = np.where(metric_val_h[i].filled(0) >= self.threshold)[0].size
            completeness = completeness / float(n_ssos)
        else:
            # The h_vals are spread more randomly among the objects
            # (we probably used one per object).
            hrange = h_vals.max() - h_vals.min()
            min_h = h_vals.min()
            if hrange < self.min_hrange:
                hrange = self.min_hrange
                min_h = h_vals.min() - hrange / 2.0
            stepsize = hrange / float(self.nbins)
            bins = np.arange(min_h, min_h + hrange + stepsize / 2.0, stepsize)
            h_vals = bins[:-1]
            n_all, b = np.histogram(metric_val_h[0], bins)
            condition = np.where(metric_val_h[0] >= self.threshold)[0]
            n_found, b = np.histogram(metric_val_h[0][condition], bins)
            completeness = n_found.astype(float) / n_all.astype(float)
            completeness = np.where(n_all == 0, 0, completeness)
        if self.cumulative:
            completeness_int = integrate_over_h(completeness, h_vals, power_law_dndh, Hindex=self.hindex)
            summary_val = np.empty(len(completeness_int), dtype=[("name", np.str_, 20), ("value", float)])
            summary_val["value"] = completeness_int
            for i, Hval in enumerate(h_vals):
                summary_val["name"][i] = "H <= %f" % (Hval)
        else:
            summary_val = np.empty(len(completeness), dtype=[("name", np.str_, 20), ("value", float)])
            summary_val["value"] = completeness
            for i, Hval in enumerate(h_vals):
                summary_val["name"][i] = "H = %f" % (Hval)
        return summary_val


class MoCompletenessAtTimeMetric(BaseMoMetric):
    """Calculate the completeness (relative to the entire population)
    <= a given H as a function of time, given the times of each discovery.

    Input values of the discovery times can come from the Discovery_Time
    (child) metric or the KnownObjects metric.

    Parameters
    ----------
    times : `np.ndarray`, (N,) or `list` [`float`]
        The bins to distribute the discovery times into.
        Same units as the discovery time (typically MJD).
    hval : `float`, optional
        The value of H to count completeness at, or cumulative completeness to.
        Default None, in which case a value halfway through h_vals
        (the slicer H range) will be chosen.
    cumulative : `bool`, optional
        If True, calculate the cumulative completeness (completeness <= H).
        If False, calculate the differential completeness (completeness @ H).
        Default None which becomes 'True',
        unless metric_name starts with 'differential'.
    hindex : `float`, optional
        Use hindex as the power law to integrate over H,
        if cumulative is True.
    """

    def __init__(self, times, hval=None, cumulative=None, hindex=0.33, **kwargs):
        self.hval = hval
        self.times = times
        self.hindex = hindex
        if cumulative is None:
            # if metric_name does not start with 'differential',
            # then cumulative->True
            if "metric_name" not in kwargs:
                self.cumulative = True
                metric_name = "CumulativeCompleteness@Time@H=%.2f" % self.hval
            else:
                #  'metric_name' in kwargs:
                metric_name = kwargs.pop("metric_name")
                if metric_name.lower().startswith("differential"):
                    self.cumulative = False
                else:
                    self.cumulative = True
        else:
            # cumulative was set
            self.cumulative = cumulative
            if "metric_name" in kwargs:
                metric_name = kwargs.pop("metric_name")
                if metric_name.lower().startswith("differential") and self.cumulative:
                    warnings.warn(f"Completeness metric_name is {metric_name} but cumulative is True")
            else:
                if self.cumulative:
                    metric_name = "CumulativeCompleteness@Time@H=%.2f" % self.hval
                else:
                    metric_name = "DifferentialCompleteness@Time@H=%.2f" % self.hval
        self._set_labels()
        super().__init__(metric_name=metric_name, units=self.units, **kwargs)

    def _set_labels(self):
        if self.hval is not None:
            if self.cumulative:
                self.units = "H <=%.1f" % (self.hval)
            else:
                self.units = "H = %.1f" % (self.hval)
        else:
            self.units = "H"

    def run(self, discovery_times, h_vals):
        if len(h_vals) != discovery_times.shape[1]:
            warnings.warn("This summary metric expects cloned H distribution. Cannot calculate summary.")
            return
        n_ssos = discovery_times.shape[0]
        timesin_h = discovery_times.swapaxes(0, 1)
        completeness_h = np.empty([len(h_vals), len(self.times)], float)
        for i, H in enumerate(h_vals):
            n, b = np.histogram(timesin_h[i].compressed(), bins=self.times)
            completeness_h[i][0] = 0
            completeness_h[i][1:] = n.cumsum()
        completeness_h = completeness_h / float(n_ssos)
        completeness = completeness_h.swapaxes(0, 1)
        if self.cumulative:
            for i, t in enumerate(self.times):
                completeness[i] = integrate_over_h(completeness[i], h_vals)
        # To save the summary statistic, we must pick out a given H value.
        if self.hval is None:
            hidx = len(h_vals) // 2
            self.hval = h_vals[hidx]
            self._set_labels()
        else:
            hidx = np.where(np.abs(h_vals - self.hval) == np.abs(h_vals - self.hval).min())[0][0]
            self.hval = h_vals[hidx]
            self._set_labels()
        summary_val = np.empty(len(self.times), dtype=[("name", np.str_, 20), ("value", float)])
        summary_val["value"] = completeness[:, hidx]
        for i, time in enumerate(self.times):
            summary_val["name"][i] = "%s @ %.2f" % (self.units, time)
        return summary_val



__all__ = ("TrackingDb", "add_run_to_database")

import os

from sqlalchemy import Column, Integer, String, create_engine
from sqlalchemy.engine import url
from sqlalchemy.exc import DatabaseError
from sqlalchemy.orm import declarative_base, sessionmaker

from . import ResultsDb, VersionRow

Base = declarative_base()


class RunRow(Base):
    """
    Define contents and format of run list table.

    Table to list all available MAF results,
    along with their opsim run and some comment info.
    """

    __tablename__ = "runs"
    # Define columns in metric list table.
    maf_run_id = Column(Integer, primary_key=True)
    run_group = Column(String)
    run_name = Column(String)
    run_comment = Column(String)
    run_version = Column(String)
    run_date = Column(String)
    db_file = Column(String)
    maf_comment = Column(String)
    maf_version = Column(String)
    maf_date = Column(String)
    maf_dir = Column(String)

    def __repr__(self):
        rstr = (
            "<Run(maf_run_id='%d', run_group='%s', run_name='%s', run_comment='%s', "
            "run_version='%s', run_date='%s', maf_comment='%s', "
            "maf_version='%s', maf_date='%s', maf_dir='%s', db_file='%s'>"
            % (
                self.maf_run_id,
                self.run_group,
                self.run_name,
                self.maf_comment,
                self.maf_version,
                self.run_date,
                self.maf_comment,
                self.maf_version,
                self.maf_date,
                self.maf_dir,
                self.db_file,
            )
        )
        return rstr


class TrackingDb:
    """Sqlite database to track MAF output runs and their locations,
    for show_maf
    """

    def __init__(self, database=None, trackingDbverbose=False):
        """
        Set up the Tracking Database.
        """
        self.verbose = trackingDbverbose
        self.driver = "sqlite"
        # connecting to non-existent database creates it automatically
        if database is None:
            # Default is a file in the current directory.
            self.database = os.path.join(os.getcwd(), "trackingDb_sqlite.db")
        else:
            self.database = database
        # only sqlite
        dbAddress = url.URL.create(drivername=self.driver, database=self.database)
        engine = create_engine(dbAddress, echo=self.verbose)
        if self.verbose:
            print("Created or connected to MAF tracking %s database at %s" % (self.driver, self.database))
        self.Session = sessionmaker(bind=engine)
        self.session = self.Session()
        # Create the tables, if they don't already exist.
        try:
            Base.metadata.create_all(engine)
        except DatabaseError:
            raise DatabaseError(
                "Cannot create a %s database at %s. Check directory exists." % (self.driver, self.database)
            )

    def close(self):
        self.session.close()

    def add_run(
        self,
        run_group=None,
        run_name=None,
        run_comment=None,
        run_version=None,
        run_date=None,
        maf_comment=None,
        maf_version=None,
        maf_date=None,
        maf_dir=None,
        db_file=None,
        maf_run_id=None,
    ):
        """Add a run to the tracking database.

        Parameters
        ----------
        run_group : `str`, optional
            Set a name to group this run with (eg. "Tier 1, 2016").
        run_name : `str`, optional
            Set a name for the opsim run.
        run_comment : `str`, optional
            Set a comment describing the opsim run.
        run_version : `str`, optional
            Set the version of opsim.
        run_date : `str`, optional
            Set the date the opsim run was created.
        maf_comment : `str`, optional
            Set a comment to describe the MAF analysis.
        maf_version : `str`, optional
            Set the version of MAF used for analysis.
        maf_date : `str`, optional
            Set the date the MAF analysis was run.
        maf_dir : `str`, optional
            The relative path to the MAF directory.
        db_file : `str`, optional
            The relative path to the Opsim SQLite database file.
        maf_run_id : `int`, optional
            The maf_run_id to assign to this record in the database
            (note this is a primary key!).
            If this run (ie the maf_dir) exists in the database already,
            this will be ignored.

        Returns
        -------
        maf_run_id : `int`
            The maf_run_id stored in the database.
        """
        if run_group is None:
            run_group = "NULL"
        if run_name is None:
            run_name = "NULL"
        if run_comment is None:
            run_comment = "NULL"
        if run_version is None:
            run_version = "NULL"
        if run_date is None:
            run_date = "NULL"
        if maf_comment is None:
            maf_comment = "NULL"
        if maf_version is None:
            maf_version = "NULL"
        if maf_date is None:
            maf_date = "NULL"
        if maf_dir is None:
            maf_dir = "NULL"
        if db_file is None:
            db_file = "NULL"
        # Test if maf_dir already exists in database.
        prevrun = self.session.query(RunRow).filter_by(maf_dir=maf_dir).all()
        if len(prevrun) > 0:
            runIds = []
            for run in prevrun:
                runIds.append(run.maf_run_id)
            print(
                "Updating information in tracking database - %s already present with runId %s."
                % (maf_dir, runIds)
            )
            for run in prevrun:
                self.session.delete(run)
            self.session.commit()
            runinfo = RunRow(
                maf_run_id=runIds[0],
                run_group=run_group,
                run_name=run_name,
                run_comment=run_comment,
                run_version=run_version,
                run_date=run_date,
                maf_comment=maf_comment,
                maf_version=maf_version,
                maf_date=maf_date,
                maf_dir=maf_dir,
                db_file=db_file,
            )
        else:
            if maf_run_id is not None:
                # Check if maf_run_id exists already.
                existing = self.session.query(RunRow).filter_by(maf_run_id=maf_run_id).all()
                if len(existing) > 0:
                    raise ValueError(
                        "maf_run_id %d already exists in database, for %s. "
                        "Record must be deleted first." % (maf_run_id, existing[0].maf_dir)
                    )
                runinfo = RunRow(
                    maf_run_id=maf_run_id,
                    run_group=run_group,
                    run_name=run_name,
                    run_comment=run_comment,
                    run_version=run_version,
                    run_date=run_date,
                    maf_comment=maf_comment,
                    maf_version=maf_version,
                    maf_date=maf_date,
                    maf_dir=maf_dir,
                    db_file=db_file,
                )
            else:
                runinfo = RunRow(
                    run_group=run_group,
                    run_name=run_name,
                    run_comment=run_comment,
                    run_version=run_version,
                    run_date=run_date,
                    maf_comment=maf_comment,
                    maf_version=maf_version,
                    maf_date=maf_date,
                    maf_dir=maf_dir,
                    db_file=db_file,
                )
        self.session.add(runinfo)
        self.session.commit()
        return runinfo.maf_run_id

    def delRun(self, runId):
        """
        Remove a run from the tracking database.
        """
        runinfo = self.session.query(RunRow).filter_by(maf_run_id=runId).all()
        if len(runinfo) == 0:
            raise Exception("Could not find run with maf_run_id %d" % (runId))
        if len(runinfo) > 1:
            raise Exception("Found more than one run with maf_run_id %d" % (runId))
        print("Removing run info for runId %d " % (runId))
        print(" ", runinfo)
        self.session.delete(runinfo[0])
        self.session.commit()


def add_run_to_database(
    maf_dir,
    tracking_db_file,
    run_group=None,
    run_name=None,
    run_comment=None,
    maf_comment=None,
    db_file=None,
    maf_version=None,
    maf_date=None,
    sched_version=None,
    sched_date=None,
    skip_extras=False,
):
    """Adds information about a MAF analysis run to a MAF tracking database.

    Parameters
    ----------
    maf_dir : `str`
        Path to the directory where the MAF results are located.
    tracking_db_file : `str`
        Full filename (+path) to the tracking database to use.
    run_group: `str`, optional
        Name to use to group this run with other opsim runs.
    run_name : `str`, optional
        Name of the opsim run.
    run_comment : `str`, optional
        Comment about the opsim run.
    run_version : `str`, optional
        Value to use for the opsim version information.
    maf_comment : `str`, optional
        Comment about the MAF analysis.
    db_file : `str`, optional
        Relative path + name of the opsim database file.
    """
    maf_dir = os.path.abspath(maf_dir)
    if not os.path.isdir(maf_dir):
        raise ValueError("There is no directory containing MAF outputs at %s." % (maf_dir))

    trackingDb = TrackingDb(database=tracking_db_file)

    # Connect to resultsDb for additional information if available
    if os.path.isfile(os.path.join(maf_dir, "resultsDb_sqlite.db")) and not skip_extras:
        resdb = ResultsDb(maf_dir)
        if run_name is None:
            run_name = resdb.get_run_name()
            if len(run_name) > 1:
                run_name = 0
            else:
                run_name = run_name[0]
        if maf_version is None or maf_date is None:
            resdb.open()
            query = resdb.session.query(VersionRow).all()
            for v in query:
                if maf_version is None:
                    maf_version = v.version
                if maf_date is None:
                    maf_date = v.run_date
            resdb.close()

    print("Adding to tracking database at %s:" % (tracking_db_file))
    print(" Maf_dir = %s" % (maf_dir))
    print(" Maf_comment = %s" % (maf_comment))
    print(" run_group = %s" % (run_group))
    print(" run_name = %s" % (run_name))
    print(" run_comment = %s" % (run_comment))
    print(" run_version = %s" % (sched_version))
    print(" run_date = %s" % (sched_date))
    print(" maf_version = %s" % (maf_version))
    print(" maf_date = %s" % (maf_date))
    print(" db_file = %s" % (db_file))
    runId = trackingDb.add_run(
        run_group=run_group,
        run_name=run_name,
        run_comment=run_comment,
        run_version=sched_version,
        run_date=sched_date,
        maf_comment=maf_comment,
        maf_version=maf_version,
        maf_date=maf_date,
        maf_dir=maf_dir,
        db_file=db_file,
    )
    print("Used MAF RunID %d" % (runId))
    trackingDb.close()



__all__ = (
    "gp_priority_map_components_to_keys",
    "galplane_priority_map",
    "GalacticPlanePriorityMap",
)

import os
import warnings

import healpy as hp
import numpy as np
from astropy.io import fits
from rubin_scheduler.data import get_data_dir

from rubin_sim.maf.maps import BaseMap
from rubin_sim.maf.utils import radec2pix


def gp_priority_map_components_to_keys(filtername, science_map):
    """A convenience function to make keeping the map key formats in sync in various places"""
    return f"galplane_priority_{science_map}:{filtername}"


def gp_priority_map_keys_to_components(mapname):
    s, f = mapname.replace("galplane_priority_", "").split(":")
    return f, s


def galplane_priority_map(
    nside=64,
    get_keys=False,
    ra=None,
    dec=None,
    pixels=None,
    interp=False,
    map_path=None,
    use_alt_maps=False,
):
    """Reads and saves the galactic plane priority maps.

    Parameters
    ----------
    nside: `int`
        Healpixel resolution (2^x). At present, this must be 64.
    get_keys : `bool`, opt
        Set this to True to retrieve *only* the keys (such as the science map names) for the maps.
        Default False.
    ra : `np.ndarray` or `float`, opt
        RA (can take numpy array). Default None sets up healpix array of nside. Radians.
    dec : `np.ndarray` or `float`, opt
        Dec (can take numpy array). Default None set up healpix array of nside. Radians.
    pixels : `np.ndarray`, opt
        Healpixel IDs, to sub-select particular healpix points. Default uses all points.
        Easiest way to access healpix values.
        Note that the pixels in the healpix array MUST come from a healpix grid with the same nside
        as the galactic plane priority map. Using different nsides can potentially fail silently.
    interp : `bool`, opt
        Should returned values be interpolated (True) or just nearest neighbor (False).
        Default False.
    map_path : `str`, opt
        Path to directory containing dust map files. Default None, uses $RUBIN_SIM_DATA_DIR/maps.
    use_alt_maps : `bool`, opt
        Use the priority_GalPlane_footprint_alt_map_data_{ugrizysum}.fits files instead of the default
        priority_galPlane_footprint_map_data_{ugrizysum}.fits files.  Default False.
    """
    # This is a function that will read the galactic plane priority map data and hold onto it indefinitely
    # this also lets us use a range of slicers, as it will set the slice_point data appropriately.

    # This function's primary goal is to return this information to the map, to use for the slicer.
    # So you MUST specify ra/dec or pixels -- or only retireve the keys
    if get_keys is False:
        if (ra is None) & (dec is None) & (pixels is None):
            raise RuntimeError("Need to set ra,dec or pixels.")

    # This reads and stores the galactic plane priority maps
    # The galactic plane priority maps are only available in nside 64
    # There are several different versions of the map - but we will almost always
    # run all of the galactic plane metrics together, so we'll just read them all at once here
    if nside != 64:
        raise RuntimeError("Currently only available with nside=64")

    # Do we need to read from disk?
    if not hasattr(galplane_priority_map, "maps"):
        galplane_priority_map.nside = 64
        galplane_priority_map.filterlist = ["u", "g", "r", "i", "z", "y", "sum"]
        if map_path is not None:
            data_dir = map_path
        else:
            data_dir = os.path.join(get_data_dir(), "maps", "GalacticPlanePriorityMaps")
        galplane_priority_map.maps = {}
        science_maps = []
        for f in galplane_priority_map.filterlist:
            if use_alt_maps:
                filename = f"priority_GalPlane_footprint_alt_map_data_{f}.fits"
            else:
                filename = f"priority_GalPlane_footprint_map_data_{f}.fits"
            mapfile = os.path.join(data_dir, filename)
            with fits.open(mapfile) as hdul:
                galplane_priority_map.maps[f] = hdul[1].data
                print(f"Read galplane priority map from {mapfile}")
                science_maps += list(galplane_priority_map.maps[f].dtype.names)
            science_maps = list(set(science_maps))
            galplane_priority_map.science_maps = science_maps

    galplane_priority_map.keys = [
        gp_priority_map_components_to_keys(f, s)
        for f in galplane_priority_map.filterlist
        for s in galplane_priority_map.science_maps
    ]

    if get_keys:
        return galplane_priority_map.keys

    if pixels is not None:
        # Assume if pixels is set, then interp is irrelevant
        maps = {}
        for k in galplane_priority_map.keys:
            f, s = gp_priority_map_keys_to_components(k)
            maps[k] = galplane_priority_map.maps[f][s][pixels]
    # BUT, if we were provided ra/dec then have to see if we should interpolate
    else:
        if interp:  # find interpolated values at given ra/dec
            maps = {}
            for k in galplane_priority_map.keys:
                f, s = gp_priority_map_keys_to_components(k)
                maps[k] = hp.get_interp_val(galplane_priority_map.maps[f][s], np.pi / 2.0 - dec, ra)
        else:  # just find nearest neighbor pixel value and return those
            pixels = radec2pix(nside, ra, dec)
            maps = {}
            for k in galplane_priority_map.keys:
                f, s = gp_priority_map_keys_to_components(k)
                maps[k] = galplane_priority_map.maps[f][s][pixels]
    return maps


class GalacticPlanePriorityMap(BaseMap):
    """
    Read and return the galactic plane priority map data at each slice_point.

    Primarily, this calls galactic_plane_priority_map to read the map data, and then assigns
    the appropriate values to each slice_point. If the slicer is an nside=64 healpix slicer, this is trivial.
    (other use-cases currently experimental and not supported).

    Parameters
    ----------
    interp : `bool`, opt
        Interpolate the dust map at each slice_point (True) or just use the nearest value (False).
        Default is False.
    nside : `int`, opt
        Default nside value to read the dust map from disk. Primarily useful if the slicer is not
        a healpix slicer.
        Default 64.
    map_path : `str`, opt
        Define a path to the directory holding the dust map files.
        Default None, which uses RUBIN_SIM_DATA_DIR.
    """

    def __init__(self, interp=False, nside=64, map_path=None):
        """
        interp: should the dust map be interpolated (True) or just use the nearest value (False).
        """
        self.keynames = galplane_priority_map(get_keys=True)
        self.interp = interp
        self.nside = nside
        self.map_path = map_path

    def run(self, slice_points):
        # If the slicer has nside, it's a healpix slicer so we can read the map directly
        if "nside" in slice_points:
            if slice_points["nside"] != self.nside:
                warnings.warn(
                    f"Slicer value of nside {slice_points['nside']} different "
                    f"from map value {self.nside}, using slicer value"
                )
            maps = galplane_priority_map(
                slice_points["nside"],
                pixels=slice_points["sid"],
                map_path=self.map_path,
            )
            for key in self.keynames:
                slice_points[key] = maps[key]
        # Not a healpix slicer, look up values based on RA,dec with possible interpolation
        else:
            maps = galplane_priority_map(
                self.nside,
                ra=slice_points["ra"],
                dec=slice_points["dec"],
                interp=self.interp,
                map_path=self.map_path,
            )
            for key in self.keynames:
                slice_points[key] = maps[key]

        return slice_points



import glob
import os

import numpy as np

from rubin_sim.data import get_data_dir
from rubin_sim.moving_objects import DirectObs, Orbits

if __name__ == "__main__":
    mjd_start = 60676.0
    length = 365.25 * 12  # How long to pre-compute for
    dtime = 1
    mjds = np.arange(mjd_start, mjd_start + length, dtime)

    orbit_files = glob.glob(os.path.join(get_data_dir(), "orbits/") + "*.txt")
    output_dir = os.path.join(os.path.join(get_data_dir(), "orbits_precompute/"))

    names = ["ra", "dec"]
    types = [float] * 2
    dt = list(zip(names, types))

    for filename in orbit_files:
        print("working on %s" % filename)
        orbits = Orbits()
        orbits.read_orbits(filename)
        # Array to hold results
        results = np.zeros((len(orbits.orbits), np.size(mjds)), dt)
        do = DirectObs()
        _temp_positions = do.generate_ephemerides(orbits, mjds, eph_mode="nbody", eph_type="basic")
        results["ra"] += _temp_positions["ra"]
        results["dec"] += _temp_positions["dec"]
        np.savez(
            os.path.join(output_dir, os.path.basename(filename).replace(".txt", ".npz")),
            positions=results,
            mjds=mjds,
        )
